# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_Bridges -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
# :license: 3-Clause BSD, see LICENSE for licensing information

"""This module has low-level functionality for parsing bridges and arranging
them into hashrings for distributors.
"""

import bisect
import logging
import re
import hashlib
import socket
import time
import ipaddr
import random

import bridgedb.Storage
import bridgedb.Bucket

from bridgedb.bridges import Bridge
from bridgedb.crypto import getHMACFunc
from bridgedb.parse import addr
from bridgedb.parse.fingerprint import toHex
from bridgedb.parse.fingerprint import fromHex
from bridgedb.parse.fingerprint import isValidFingerprint
from bridgedb.safelog import logSafely

try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO


ID_LEN = 20  # XXX Only used in commented out line in Storage.py
DIGEST_LEN = 20
PORTSPEC_LEN = 16

re_ipv6 = re.compile("\[([a-fA-F0-9:]+)\]:(.*$)")
re_ipv4 = re.compile("((?:\d{1,3}\.?){4}):(.*$)")


def parseCountryBlockFile(f):
    """Generator. Parses a blocked-bridges file 'f', and yields
       a fingerprint (ID), address, a list of ports, and a list of country
       codes where the bridge is blocked for each valid line:
       address, port [], countrycode []"""
    for line in f:
        ID = address = fields = portlist = countries = None
        line = line.strip()
        try:
            ID, addrspec, countries = line.split()
            if isValidFingerprint(ID):
                ID = fromHex(ID)
                logging.debug("Parsed ID: %s", ID)
            else:
                print "failed to parse ID!"
                continue # skip this line

            for regex in [re_ipv4, re_ipv6]:
                m = regex.match(addrspec)
                if m:
                    address = ipaddr.IPAddress(m.group(1))
                    portlist = addr.PortList(m.group(2))
                    countries = countries.split(',')
                    logging.debug("Parsed address: %s", address)
                    logging.debug("Parsed portlist: %s", portlist)
                    logging.debug("Parsed countries: %s", countries)
        except (IndexError, ValueError):
            logging.debug("Skipping line")
            continue # skip this line
        if ID and address and portlist and countries:
            yield ID, address, portlist, countries

class BridgeHolder(object):
    """Abstract base class for all classes that hold bridges."""
    def insert(self, bridge):
        raise NotImplementedError

    def clear(self):
        pass

    def dumpAssignments(self, f, description=""):
        pass


class BridgeRingParameters(object):
    """Store validated settings on minimum number of Bridges with certain
    attributes which should be included in any generated subring of a
    hashring.

    :ivar list needPorts: List of two-tuples of desired port numbers and their
        respective minimums.
    :ivar list needFlags: List of two-tuples of desired flags_ assigned to a
        Bridge by the Bridge DirAuth.

    .. _flags: https://gitweb.torproject.org/torspec.git/blob/HEAD:/dir-spec.txt#l1696
    """

    def __init__(self, needPorts=[], needFlags=[]):
        """Control the creation of subrings by including a minimum number of
        bridges which possess certain attributes.

        :type needPorts: iterable
        :param needPorts: An iterable of two-tuples. Each two tuple should
            contain ``(port, minimum)``, where ``port`` is an integer
            specifying a port number, and ``minimum`` is another integer
            specifying the minimum number of Bridges running on that ``port``
            to include in any new subring.
        :type needFlags: iterable
        :param needFlags: An iterable of two-tuples. Each two tuple should
            contain ``(flag, minimum)``, where ``flag`` is a string specifying
            an OR flag_, and ``minimum`` is an integer for the minimum number
            of Bridges which have acquired that ``flag`` to include in any new
            subring.
        :raises: An :exc:`TypeError` if an invalid port number, a minimum less
            than one, or an "unsupported" flag is given. "Stable" appears to
            be the only currently "supported" flag.
        """
        for port, count in needPorts:
            if not (1 <= port <= 65535):
                raise TypeError("Port %s out of range." % port)
            if count <= 0:
                raise TypeError("Count %s out of range." % count)
        for flag, count in needFlags:
            flag = flag.lower()
            if flag not in ["stable", "running",]:
                raise TypeError("Unsupported flag %s" % flag)
            if count <= 0:
                raise TypeError("Count %s out of range." % count)

        self.needPorts = needPorts[:]
        self.needFlags = [(flag.lower(), count) for flag, count in needFlags[:]]

class BridgeRing(BridgeHolder):
    """Arranges bridges into a hashring based on an hmac function."""

    def __init__(self, key, answerParameters=None):
        """Create a new BridgeRing, using key as its hmac key.

        :type key: bytes
        :param key: The HMAC key, generated with
                    :func:`bridgedb.crypto.getKey`.
        :type answerParameters: :class:`BridgeRingParameters`
        :param answerParameters: DOCDOC
        :ivar dict bridges: A dictionary which maps HMAC keys to
                            :class:`~bridgedb.Bridges.Bridge`s.
        :ivar dict bridgesByID: A dictionary which maps raw hash digests of
                                bridge ID keys to
                                :class:`~bridgedb.Bridges.Bridge`s.
        :type hmac: callable
        :ivar hmac: An HMAC function, which uses the **key** parameter to
                    generate new HMACs for storing, inserting, and retrieving
                    :class:`~bridgedb.Bridges.Bridge`s within mappings.
        :ivar bool isSorted: ``True`` if ``sortedKeys`` is currently sorted.
        :ivar list sortedKeys: A sorted list of all of the HMACs.
        :ivar str name: A string which identifies this hashring, used mostly
                        for differentiating this hashring in log messages, but
                        it is also used for naming subrings. If this hashring
                        is a subring, the ``name`` will include whatever
                        distinguishing parameters differentiate that
                        particular subring (i.e. ``'(port-443 subring)'`` or
                        ``'(Stable subring)'``)
        :type subrings: list
        :ivar subrings: A list of other ``BridgeRing``s, each of which
                        contains bridges of a particular type. For example, a
                        subring might contain only ``Bridge``s which have been
                        given the "Stable" flag, or it might contain only IPv6
                        bridges. Each item in this list should be a 4-tuple:

                          ``(type, value, count, ring)``

                        where:

                          * ``type`` is a string which describes what kind of
                            parameter is used to determine if a ``Bridge``
                            belongs in that subring, i.e. ``'port'`` or
                            ``'flag'``.

                          * ``value`` is a specific value pertaining to the
                            ``type``, e.g. ``type='port'; value=443``.

                          * ``count`` is an integer for the current total
                             number of bridges in the subring.

                          * ``ring`` is a
                            :class:`~bridgedb.Bridges.BridgeRing`; it is the
                            sub hashring which contains ``count`` number of
                            :class:`~bridgedb.Bridges.Bridge`s of a certain
                            ``type``.
        """
        self.bridges = {}
        self.bridgesByID = {}
        self.hmac = getHMACFunc(key, hex=False)
        self.isSorted = False
        self.sortedKeys = []
        if answerParameters is None:
            answerParameters = BridgeRingParameters()
        self.answerParameters = answerParameters

        self.subrings = []
        for port,count in self.answerParameters.needPorts:
            #note that we really need to use the same key here, so that
            # the mapping is in the same order for all subrings.
            self.subrings.append( ('port',port,count,BridgeRing(key,None)) )
        for flag,count in self.answerParameters.needFlags:
            self.subrings.append( ('flag',flag,count,BridgeRing(key,None)) )

        self.setName("Ring")

    def setName(self, name):
        """Tag a unique name to this hashring for identification.

        :param string name: The name for this hashring.
        """
        self.name = name
        for tp, val, _, subring in self.subrings:
            if tp == 'port':
                subring.setName("%s (port-%s subring)" % (name, val))
            else:
                subring.setName("%s (%s subring)" % (name, val))

    def __len__(self):
        """Get the number of unique bridges this hashring contains."""
        return len(self.bridges)

    def clear(self):
        """Remove all bridges and mappings from this hashring and subrings."""
        self.bridges = {}
        self.bridgesByID = {}
        self.isSorted = False
        self.sortedKeys = []

        for tp, val, count, subring in self.subrings:
            subring.clear()

    def insert(self, bridge):
        """Add a **bridge** to this hashring.

        The bridge's position in the hashring is dependent upon the HMAC of
        the raw hash digest of the bridge's ID key. The function used to
        generate the HMAC, :ivar:`BridgeRing.hmac`, is unique to each
        individual hashring.

        If the (presumably same) bridge is already at that determined position
        in this hashring, replace the old one.

        :type bridge: :class:`~bridgedb.Bridges.Bridge`
        :param bridge: The bridge to insert into this hashring.
        """
        for tp, val, _, subring in self.subrings:
            if tp == 'port':
                if val == bridge.orPort:
                    subring.insert(bridge)
            else:
                assert tp == 'flag' and val == 'stable'
                if val == 'stable' and bridge.flags.stable:
                    subring.insert(bridge)

        ident = bridge.getID()
        pos = self.hmac(ident)
        if not self.bridges.has_key(pos):
            self.sortedKeys.append(pos)
            self.isSorted = False
        self.bridges[pos] = bridge
        self.bridgesByID[ident] = bridge
        logging.debug("Adding %s to %s" % (bridge.address, self.name))

    def _sort(self):
        """Helper: put the keys in sorted order."""
        if not self.isSorted:
            self.sortedKeys.sort()
            self.isSorted = True

    def _getBridgeKeysAt(self, pos, N=1):
        """Bisect a list of bridges at a specified position, **pos**, and
        retrieve bridges from that point onwards, wrapping around the hashring
        if necessary.

        If the number of bridges requested, **N**, is larger that the size of
        this hashring, return the entire ring. Otherwise:

          1. Sort this bridges in this hashring, if it is currently unsorted.

          2. Bisect the sorted bridges. If the bridge at the desired position,
             **pos**, already exists within this hashring, the the bisection
             result is the bridge at position **pos**. Otherwise, the bisection
             result is the first position after **pos** which has a bridge
             assigned to it.

          3. Try to obtain **N** bridges, starting at (and including) the
             bridge in the requested position, **pos**.

               a. If there aren't **N** bridges after **pos**, wrap back
                  around to the beginning of the hashring and obtain bridges
                  until we have **N** bridges.

          4. Check that the number of bridges obtained is indeed **N**, then
             return them.

        :param bytes pos: The position to jump to. Any bridges returned will
                          start at this position in the hashring, if there is
                          a bridge assigned to that position. Otherwise,
                          indexing will start at the first position after this
                          one which has a bridge assigned to it.
        :param int N: The number of bridges to return.
        :rtype: list
        :returns: A list of :class:`~bridgedb.Bridges.Bridge`s.
        """
        assert len(pos) == DIGEST_LEN
        if N >= len(self.sortedKeys):
            return self.sortedKeys
        if not self.isSorted:
            self._sort()
        idx = bisect.bisect_left(self.sortedKeys, pos)
        r = self.sortedKeys[idx:idx+N]
        if len(r) < N:
            # wrap around as needed.
            r.extend(self.sortedKeys[:N - len(r)])
        assert len(r) == N
        return r

    def getBridges(self, pos, N=1, countryCode=None):
        """Return **N** bridges appearing in this hashring after a position.

        :param bytes pos: The position to jump to. Any bridges returned will
                          start at this position in the hashring, if there is
                          a bridge assigned to that position. Otherwise,
                          indexing will start at the first position after this
                          one which has a bridge assigned to it.
        :param int N: The number of bridges to return.
        :type countryCode: str or None
        :param countryCode: DOCDOC
        :rtype: list
        :returns: A list of :class:`~bridgedb.Bridges.Bridge`s.
        """
        # XXX This can be removed after we determine if countryCode is ever
        # actually being used. It seems the countryCode should be passed in
        # from bridgedb.HTTPServer.WebResource.getBridgeRequestAnswer() in
        # order to hand out bridges which are believed to not be blocked in a
        # given country.
        if countryCode:
            logging.debug("getBridges: countryCode=%r" % countryCode)

        forced = []
        for _, _, count, subring in self.subrings:
            if len(subring) < count:
                count = len(subring)
            forced.extend(subring._getBridgeKeysAt(pos, count))

        keys = [ ]
        for k in forced + self._getBridgeKeysAt(pos, N):
            if k not in keys:
                keys.append(k)
            else:
                logging.debug(
                    "Got duplicate bridge %r in main hashring for position %r."
                    % (logSafely(k.encode('hex')), pos.encode('hex')))
        keys = keys[:N]
        keys.sort()

        #Do not return bridges from the same /16
        bridges = [ self.bridges[k] for k in keys ]

        return bridges

    def getBridgeByID(self, fp):
        """Return the bridge whose identity digest is fp, or None if no such
           bridge exists."""
        for _,_,_,subring in self.subrings:
            b = subring.getBridgeByID(fp)
            if b is not None:
                return b

        return self.bridgesByID.get(fp)

    def dumpAssignments(self, f, description=""):
        logging.info("Dumping bridge assignments for %s..." % self.name)
        for b in self.bridges.itervalues():
            desc = [ description ]
            ident = b.getID()
            for tp,val,_,subring in self.subrings:
                if subring.getBridgeByID(ident):
                    desc.append("%s=%s"%(tp,val))
            f.write("%s %s\n"%( toHex(ident), " ".join(desc).strip()))

class FixedBridgeSplitter(BridgeHolder):
    """A bridgeholder that splits bridges up based on an hmac and assigns
       them to several sub-bridgeholders with equal probability.
    """
    def __init__(self, key, rings):
        self.hmac = getHMACFunc(key, hex=True)
        self.rings = rings[:]
        for r in self.rings:
            assert(isinstance(r, BridgeHolder))

    def insert(self, bridge):
        # Grab the first 4 bytes
        digest = self.hmac(bridge.getID())
        pos = long( digest[:8], 16 )
        which = pos % len(self.rings)
        self.rings[which].insert(bridge)

    def clear(self):
        """Clear all bridges from every ring in ``rings``."""
        for r in self.rings:
            r.clear()

    def __len__(self):
        """Returns the total number of bridges in all ``rings``."""
        total = 0
        for ring in self.rings:
            total += len(ring)
        return total

    def dumpAssignments(self, filename, description=""):
        """Write all bridges assigned to this hashring to ``filename``.

        :param string description: If given, include a description next to the
            index number of the ring from :attr:`FilteredBridgeHolder.rings`
            the following bridges were assigned to. For example, if the
            description is ``"IPv6 obfs2 bridges"`` the line would read:
            ``"IPv6 obfs2 bridges ring=3"``.
        """
        for index, ring in zip(xrange(len(self.rings)), self.rings):
            ring.dumpAssignments(filename, "%s ring=%s" % (description, index))

class UnallocatedHolder(BridgeHolder):
    """A pseudo-bridgeholder that ignores its bridges and leaves them
       unassigned.
    """
    def __init__(self):
        self.fingerprints = []

    def insert(self, bridge):
        logging.debug("Leaving %s unallocated", bridge.fingerprint)
        if not bridge.fingerprint in self.fingerprints:
            self.fingerprints.append(bridge.fingerprint)

    def __len__(self):
        return len(self.fingerprints)

    def clear(self):
        self.fingerprints = []

    def dumpAssignments(self, f, description=""):
        with bridgedb.Storage.getDB() as db:
            allBridges = db.getAllBridges()
            for bridge in allBridges:
                if bridge.hex_key not in self.fingerprints:
                    continue
                dist = bridge.distributor
                desc = [ description ]
                if dist.startswith(bridgedb.Bucket.PSEUDO_DISTRI_PREFIX):
                    dist = dist.replace(bridgedb.Bucket.PSEUDO_DISTRI_PREFIX, "")
                    desc.append("bucket=%s" % dist)
                elif dist != "unallocated":
                    continue
                f.write("%s %s\n" % (bridge.hex_key, " ".join(desc).strip()))

class BridgeSplitter(BridgeHolder):
    """A BridgeHolder that splits incoming bridges up based on an hmac,
       and assigns them to sub-bridgeholders with different probabilities.
       Bridge-to-bridgeholder associations are recorded in a store.
    """
    def __init__(self, key):
        self.hmac = getHMACFunc(key, hex=True)
        self.ringsByName = {}
        self.totalP = 0
        self.pValues = []
        self.rings = []
        self.pseudoRings = []
        self.statsHolders = []

    def __len__(self):
        n = 0
        for r in self.ringsByName.values():
            n += len(r)
        return n

    def addRing(self, ring, ringname, p=1):
        """Add a new bridgeholder.
           ring -- the bridgeholder to add.
           ringname -- a string representing the bridgeholder.  This is used
               to record which bridges have been assigned where in the store.
           p -- the relative proportion of bridges to assign to this
               bridgeholder.
        """
        assert isinstance(ring, BridgeHolder)
        self.ringsByName[ringname] = ring
        self.pValues.append(self.totalP)
        self.rings.append(ringname)
        self.totalP += p

    def addPseudoRing(self, ringname):
        """Add a pseudo ring to the list of pseudo rings.
        """
        self.pseudoRings.append(bridgedb.Bucket.PSEUDO_DISTRI_PREFIX + ringname)

    def addTracker(self, t):
        """Adds a statistics tracker that gets told about every bridge we see.
        """
        self.statsHolders.append(t)

    def clear(self):
        for r in self.ringsByName.values():
            r.clear()

    def insert(self, bridge):
        assert self.rings

        for s in self.statsHolders:
            s.insert(bridge)

        # The bridge must be running to insert it:
        if not bridge.flags.running:
            return

        bridgeID = bridge.fingerprint

        # Determine which ring to put this bridge in if we haven't seen it
        # before.
        pos = self.hmac(bridgeID)
        n = int(pos[:8], 16) % self.totalP
        pos = bisect.bisect_right(self.pValues, n) - 1
        assert 0 <= pos < len(self.rings)
        ringname = self.rings[pos]
        logging.info("%s placing bridge %s into hashring %s (via n=%s, pos=%s)."
                     % (self.__class__.__name__, bridge, ringname, n, pos))

        validRings = self.rings + self.pseudoRings

        with bridgedb.Storage.getDB() as db:
            ringname = db.insertBridgeAndGetRing(bridge, ringname, time.time(), 
                                             validRings)
            db.commit()

            # Pseudo distributors are always held in the "unallocated" ring
            if ringname in self.pseudoRings:
                ringname = "unallocated"

            ring = self.ringsByName.get(ringname)
            ring.insert(bridge)

    def dumpAssignments(self, f, description=""):
        for name,ring in self.ringsByName.iteritems():
            ring.dumpAssignments(f, "%s %s" % (description, name))


class FilteredBridgeSplitter(BridgeHolder):
    """A configurable BridgeHolder that filters bridges into subrings.

    The set of subrings and conditions used to assign :class:`Bridge`s should
    be passed to :meth:`~FilteredBridgeSplitter.addRing`.
    """

    def __init__(self, key, max_cached_rings=3):
        """Create a hashring which filters bridges into sub hashrings.

        :type key: DOCDOC
        :param key: An HMAC key.
        :param int max_cached_rings: XXX max_cached_rings appears to not be
             used anywhere.

        :ivar filterRings: A dictionary of subrings which has the form
             ``{ringname: (filterFn, subring)}``, where:
                 - ``ringname`` is a unique string identifying the subring.
                 - ``filterFn`` is a callable which filters Bridges in some
                   manner, i.e. by whether they are IPv4 or IPv6, etc.
                 - ``subring`` is a :class:`BridgeHolder`.
        :ivar hmac: DOCDOC
        :ivar bridges: DOCDOC
        :type distributorName: str
        :ivar distributorName: The name of this splitter's distributor. See
             :meth:`bridgedb.Dist.IPBasedDistributor.setDistributorName`.
        """
        self.key = key
        self.filterRings = {}
        self.hmac = getHMACFunc(key, hex=True)
        self.bridges = []
        self.distributorName = ''

        #XXX: unused
        self.max_cached_rings = max_cached_rings

    def __len__(self):
        return len(self.bridges)

    def clear(self):
        self.bridges = []
        self.filterRings = {}

    def insert(self, bridge):
        """Insert a bridge into all appropriate sub-hashrings.

        For all sub-hashrings, the ``bridge`` will only be added iff it passes
        the filter functions for that sub-hashring.

        :type bridge: :class:`~bridgedb.Bridges.Bridge`
        :param bridge: The bridge to add.
        """
        # The bridge must be running to insert it:
        if not bridge.flags.running:
            logging.warn(
                "Skipping hashring insertion for non-running bridge: '%s'"
                % logSafely(bridge.fingerprint))
            return

        index = 0
        logging.debug("Inserting %s into splitter"
                      % (logSafely(bridge.fingerprint)))
        for old_bridge in self.bridges[:]:
            if bridge.fingerprint == old_bridge.fingerprint:
                self.bridges[index] = bridge
                break
            index += 1
        else:
            self.bridges.append(bridge)
        for ringname, (filterFn, subring) in self.filterRings.items():
            if filterFn(bridge):
                subring.insert(bridge)
                logging.debug("Inserted bridge '%s' into '%s' sub hashring"
                              % (logSafely(bridge.fingerprint), ringname))

    def extractFilterNames(self, ringname):
        """Get the names of the filters applied to a particular sub hashring.

        :param str ringname: A unique name identifying a sub hashring.
        :rtype: list
        :returns: A sorted list of strings, all the function names of the
                  filters applied to the sub hashring named **ringname**.
        """
        filterNames = []

        for filterName in [x.func_name for x in list(ringname)]:
            # Using `filterAssignBridgesToRing.func_name` gives us a messy
            # string which includes all parameters and memory addresses. Get
            # rid of this by partitioning at the first `(`:
            realFilterName = filterName.partition('(')[0]
            filterNames.append(realFilterName)

        filterNames.sort()
        return filterNames

    def addRing(self, subring, ringname, filterFn, populate_from=None):
        """Add a subring to this hashring.

        :type subring: :class:`BridgeHolder`
        :param subring: The subring to add.
        :param str ringname: A unique name for identifying the new subring.
        :param filterFn: A function whose input is a :class:`Bridge`, and
                         returns True/False based on some filtration criteria.
        :type populate_from: iterable or None
        :param populate_from: A group of :class:`Bridge`s. If given, the newly
                              added subring will be populated with these
                              bridges.
        :rtype: bool
        :returns: False if there was a problem adding the subring, True
                  otherwise.
        """
        # XXX I think subring and ringname are switched in this function, or
        # at least that whatever is passed into this function as as the
        # `ringname` parameter from somewhere else is odd; for example, with
        # the original code, which was `log.debug("Inserted %d bridges into
        # hashring '%s'!" % (inserted, ringname))`, this log message appears:
        #
        # Jan 04 23:18:37 [INFO] Inserted 12 bridges into hashring
        # frozenset([<function filterBridgesByIP4 at 0x2d67cf8>, <function
        # filterAssignBridgesToRing(<function hmac_fn at 0x3778398>, 4, 0) at
        # 0x37de578>])!
        #
        # I suppose since it contains memory addresses, it *is* technically
        # likely to be a unique string, but it is messy.

        logging.debug("Adding '%s' subring to %s..."
                      % (ringname, self.__class__.__name__))

        if not isinstance(subring, BridgeHolder):
            logging.fatal("%s hashring can't add invalid subring: %r"
                          % (self.distributorName, subring))
            return False
        if ringname in self.filterRings.keys():
            logging.fatal("%s hashring already has a subring named '%s'!"
                          % (self.distributorName, ringname))
            return False

        filterNames = self.extractFilterNames(ringname)
        subringName = [self.distributorName]
        for filterName in filterNames:
            if filterName != 'filterAssignBridgesToRing':
                subringName.append(filterName.strip('filterBridgesBy'))
        subringName = '-'.join([x for x in subringName])
        subring.setName(subringName)

        logging.info("Adding subring to %s hashring..." % subring.name)
        logging.info("  Subring filters: %s" % filterNames)

        #TODO: drop LRU ring if len(self.filterRings) > self.max_cached_rings
        self.filterRings[ringname] = (filterFn, subring)

        if populate_from:
            inserted = 0
            for bridge in populate_from:
                if isinstance(bridge, Bridge) and filterFn(bridge):
                    subring.insert(bridge)
                    inserted += 1
            logging.info("Bridges inserted into %s subring: %d"
                         % (subring.name, inserted))

        return True

    def dumpAssignments(self, f, description=""):
        # one ring per filter set
        # bridges may be present in multiple filter sets
        # only one line should be dumped per bridge

        for b in self.bridges:
            # gather all the filter descriptions
            desc = []
            for n,(g,r) in self.filterRings.items():
                if g(b):
                    # ghetto. get subring flags, ports
                    for tp,val,_,subring in r.subrings:
                        if subring.getBridgeByID(b.getID()):
                            desc.append("%s=%s"%(tp,val))
                    try:
                        desc.extend(g.description.split())
                    except TypeError:
                        desc.append(g.description)

            # add transports
            logging.debug("%s supports %d transports" % (b, len(b.transports)))
            for transport in b.transports:
                desc.append("transport=%s"%(transport.methodname))

            # dedupe and group
            desc = set(desc)
            grouped = dict()
            for kw in desc:
                l,r = kw.split('=')
                try:
                    grouped[l] = "%s,%s"%(grouped[l],r)
                except KeyError:
                    grouped[l] = kw

            # add to assignments
            desc = "%s %s" % (description.strip(),
                    " ".join([v for k,v in grouped.items()]).strip())
            f.write("%s %s\n"%( toHex(b.getID()), desc))

class BridgeBlock(object):
    """Base class that abstracts bridge blocking.

    .. TODO:: This should be a zope.interface specification.
    """
    def __init__(self):
        pass

    def insert(self, fingerprint, blockingRule):
        raise NotImplementedError

    def clear(self):
        pass

class CountryBlock(BridgeBlock):
    """Countrywide bridge blocking"""
    def __init__(self):
        self.db = bridgedb.Storage.getDB()

    def clear(self):
        assert self.db
        self.db.cleanBridgeBlocks()
        self.db.commit()

    def insert(self, fingerprint, blockingRule):
        """ insert a country based blocking rule """
        assert self.db
        countryCode = blockingRule
        self.db.addBridgeBlock(fingerprint, countryCode)
        self.db.commit()

    def getBlockingCountries(self, fingerprint):
        """ returns a list of country codes where this fingerprint is blocked"""
        assert self.db
        if fingerprint is not None:
            return self.db.getBlockingCountries(fingerprint) 



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_runner -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2007-2015, all entities within the AUTHORS file
#             (c) 2012-2015, Isis Lovecruft
# :license: 3-clause BSD, see included LICENSE for information

"""Classes for running components and servers, as well as daemonisation.

** Module Overview: **

"""

from __future__ import print_function

import logging
import sys
import os

from twisted.python import procutils


def find(filename):
    """Find the executable ``filename``.

    :param string filename: The executable to search for. Must be in the
       effective user ID's $PATH.
    :rtype: string
     :returns: The location of the executable, if found. Otherwise, returns
        None.
    """
    executable = None

    logging.debug("Searching for installed '%s'..." % filename)
    which = procutils.which(filename, os.X_OK)

    if len(which) > 0:
        for that in which:
            if os.stat(that).st_uid == os.geteuid():
                executable = that
                break
    if not executable:
        return None

    logging.debug("Found installed script at '%s'" % executable)
    return executable

def generateDescriptors(count=None, rundir=None):
    """Run a script which creates fake bridge descriptors for testing purposes.

    This will run Leekspin_ to create bridge server descriptors, bridge
    extra-info descriptors, and networkstatus document.

    .. warning: This function can take a very long time to run, especially in
        headless environments where entropy sources are minimal, because it
        creates the keys for each mocked OR, which are embedded in the server
        descriptors, used to calculate the OR fingerprints, and sign the
        descriptors, among other things.

    .. _Leekspin: https://gitweb.torproject.org/user/isis/leekspin.git

    :param integer count: Number of mocked bridges to generate descriptor
        for. (default: 3)
    :type rundir: string or None
    :param rundir: If given, use this directory as the current working
        directory for the bridge descriptor generator script to run in. The
        directory MUST already exist, and the descriptor files will be created
        in it. If None, use the whatever directory we are currently in.
    """
    import subprocess
    import os.path

    proc = None
    statuscode = 0
    script = 'leekspin'
    rundir = rundir if os.path.isdir(rundir) else None
    count = count if count else 3
    try:
        proc = subprocess.Popen([script, '-n', str(count)],
                                close_fds=True, cwd=rundir)
    finally:
        if proc is not None:
            proc.wait()
            if proc.returncode:
                print("There was an error generating bridge descriptors.",
                      "(Returncode: %d)" % proc.returncode)
                statuscode = proc.returncode
            else:
                print("Sucessfully generated %s descriptors." % str(count))
        del subprocess
        return statuscode

def runTrial(options):
    """Run Twisted trial based unittests, optionally with coverage.

    :type options: :class:`~bridgedb.opt.TestOptions`
    :param options: Parsed options for controlling the twisted.trial test
        run. All unrecognised arguments after the known options will be passed
        along to trial.
    """
    from twisted.scripts import trial

    # Insert 'trial' as the first system cmdline argument:
    sys.argv = ['trial']

    if options['coverage']:
        try:
            from coverage import coverage
        except ImportError as ie:
            print(ie.message)
        else:
            cov = coverage()
            cov.start()
            sys.argv.append('--coverage')
            sys.argv.append('--reporter=bwverbose')

    # Pass all arguments along to its options parser:
    if 'test_args' in options:
        for arg in options['test_args']:
            sys.argv.append(arg)
    # Tell trial to test the bridgedb package:
    sys.argv.append('bridgedb.test')
    trial.run()

    if options['coverage']:
        cov.stop()
        cov.html_report('_trial_temp/coverage/')

def runTests(options):
    """Run unittest based tests.

    :type options: :class:`~bridgedb.opt.TestOptions`
    :param options: Parsed options for controlling the twisted.trial test
        run. All unrecognised arguments after the known options will be passed
        along to trial.
    """
    testModule = __import__('bridgedb.Tests', globals(), '', [])
    testModule.Tests.main()

def doDumpBridges(config):
    """Dump bridges by assignment to a file.

    This function handles the commandline '--dump-bridges' option.

    :type config: :class:`bridgedb.Main.Conf`
    :param config: The current configuration.
    """
    import bridgedb.Bucket as bucket

    bucketManager = bucket.BucketManager(config)
    bucketManager.assignBridgesToBuckets()
    bucketManager.dumpBridges()



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_email_dkim -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Nick Mathewson <nickm@torproject.org>
#           Isis Lovecruft <isis@torproject.org> 0xA3ADB67A2CDB8B35
#           Matthew Finkel <sysrqb@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2013-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________

"""
.. py:module:: bridgedb.email.dkim
    :synopsis: Functions for checking DKIM verification results in email
               headers.

bridgedb.email.dkim
===================

Functions for checking DKIM verification results in email headers.

::

 bridgedb.email.dkim
  |_ checkDKIM - Check the DKIM verification results header.
..
"""

from __future__ import unicode_literals

import logging


def checkDKIM(message, rules):
    """Check the DKIM verification results header.

    This check is only run if the incoming email, **message**, originated from
    a domain for which we're configured (in the ``EMAIL_DOMAIN_RULES``
    dictionary in the config file) to check DKIM verification results for.

    Returns ``False`` if:

    1. We're supposed to expect and check the DKIM headers for the
       client's email provider domain.
    2. Those headers were *not* okay.

    Otherwise, returns ``True``.

    :type message: :api:`twisted.mail.smtp.rfc822.Message`
    :param message: The incoming client request email, including headers.
    :param dict rules: The list of configured ``EMAIL_DOMAIN_RULES`` for the
        canonical domain which the client's email request originated from.
    :rtype: bool
    :returns: ``False`` if the checks failed, ``True`` otherwise.
    """
    logging.info("Checking DKIM verification results...")
    logging.debug("Domain has rules: %s" % ', '.join(rules))

    if 'dkim' in rules:
        # getheader() returns the last of a given kind of header; we want
        # to get the first, so we use getheaders() instead.
        dkimHeaders = message.getheaders("X-DKIM-Authentication-Results")
        dkimHeader = "<no header>"
        if dkimHeaders:
            dkimHeader = dkimHeaders[0]
        if not dkimHeader.startswith("pass"):
            logging.info("Rejecting bad DKIM header on incoming email: %r "
                         % dkimHeader)
            return False
    return True



# -*- encoding: utf-8 -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
# :copyright: (c) 2014-2015, Isis Lovecruft
#             (c) 2014-2015, The Tor Project, Inc.
# :license: see LICENSE for licensing information
#_____________________________________________________________________________

"""Unittests for :mod:`bridgedb.HTTPServer`."""

from __future__ import print_function

import logging
import os
import random
import shutil

import ipaddr

from BeautifulSoup import BeautifulSoup

from twisted.internet import reactor
from twisted.internet import task
from twisted.trial import unittest
from twisted.web.resource import Resource
from twisted.web.test import requesthelper

from bridgedb import HTTPServer
from bridgedb.schedule import ScheduledInterval


# For additional logger output for debugging, comment out the following:
logging.disable(50)
# and then uncomment the following line:
#HTTPServer.logging.getLogger().setLevel(10)


class ReplaceErrorPageTests(unittest.TestCase):
    """Tests for :func:`bridgedb.HTTPServer.replaceErrorPage`."""

    def test_replaceErrorPage(self):
        """``replaceErrorPage`` should return the expected html."""
        exc = Exception("vegan gümmibären")
        errorPage = HTTPServer.replaceErrorPage(exc)
        self.assertSubstring("Something went wrong", errorPage)
        self.assertNotSubstring("vegan gümmibären", errorPage)


class CaptchaProtectedResourceTests(unittest.TestCase):
    """Tests for :mod:`bridgedb.HTTPServer.CaptchaProtectedResource`."""

    def setUp(self):
        self.dist = None
        self.sched = None
        self.pagename = b'bridges.html'
        self.root = Resource()
        self.protectedResource = HTTPServer.WebResourceBridges(self.dist,
                                                               self.sched)
        self.captchaResource = HTTPServer.CaptchaProtectedResource(
            useForwardedHeader=True, protectedResource=self.protectedResource)
        self.root.putChild(self.pagename, self.captchaResource)

    def test_render_GET_noCaptcha(self):
        """render_GET() should return a page without a CAPTCHA, which has the
        image alt text.
        """
        request = DummyRequest([self.pagename])
        request.method = b'GET'
        page = self.captchaResource.render_GET(request)
        self.assertSubstring(
            "Your browser is not displaying images properly", page)

    def test_render_GET_missingTemplate(self):
        """render_GET() with a missing template should raise an error and
        return the result of replaceErrorPage().
        """
        oldLookup = HTTPServer.lookup
        try:
            HTTPServer.lookup = None
            request = DummyRequest([self.pagename])
            request.method = b'GET'
            page = self.captchaResource.render_GET(request)
            errorPage = HTTPServer.replaceErrorPage(Exception('kablam'))
            self.assertEqual(page, errorPage)
        finally:
            HTTPServer.lookup = oldLookup

    def createRequestWithIPs(self):
        """Set the IP address returned from ``request.getClientIP()`` to
        '3.3.3.3', and the IP address reported in the 'X-Forwarded-For' header
        to '2.2.2.2'.
        """
        request = DummyRequest([self.pagename])
        # Since we do not set ``request.getClientIP`` here like we do in some
        # of the other unittests, an exception would be raised here if
        # ``getBridgesForRequest()`` is unable to get the IP address from this
        # 'X-Forwarded-For' header (because ``ip`` would get set to ``None``).
        request.headers.update({'x-forwarded-for': '2.2.2.2'})
        # See :api:`twisted.test.requesthelper.DummyRequest.getClientIP`
        request.client = requesthelper.IPv4Address('TCP', '3.3.3.3', 443)
        request.method = b'GET'
        return request

    def test_getClientIP_XForwardedFor(self):
        """CaptchaProtectedResource.getClientIP() should return the IP address
        from the 'X-Forwarded-For' header when ``useForwardedHeader=True``.
        """
        self.captchaResource.useForwardedHeader = True
        request = self.createRequestWithIPs()
        clientIP = self.captchaResource.getClientIP(request)
        self.assertEqual(clientIP, '2.2.2.2')

    def test_getClientIP_fromRequest(self):
        """CaptchaProtectedResource.getClientIP() should return the IP address
        from the request instance when ``useForwardedHeader=False``.
        """
        self.captchaResource.useForwardedHeader = False
        request = self.createRequestWithIPs()
        clientIP = self.captchaResource.getClientIP(request)
        self.assertEqual(clientIP, '3.3.3.3')

    def test_render_POST(self):
        """render_POST() with a wrong 'captcha_response_field' should return
        a redirect to the CaptchaProtectedResource page.
        """
        request = DummyRequest([self.pagename])
        request.method = b'POST'
        page = self.captchaResource.render_POST(request)
        self.assertEqual(BeautifulSoup(page).find('meta')['http-equiv'],
                         'refresh')


class GimpCaptchaProtectedResourceTests(unittest.TestCase):
    """Tests for :mod:`bridgedb.HTTPServer.GimpCaptchaProtectedResource`."""

    def setUp(self):
        """Create a :class:`HTTPServer.WebResourceBridges` and protect it with
        a :class:`GimpCaptchaProtectedResource`.
        """
        # Create our cached CAPTCHA directory:
        self.captchaDir = 'captchas'
        if not os.path.isdir(self.captchaDir):
            os.makedirs(self.captchaDir)

        # Set up our resources to fake a minimal HTTP(S) server:
        self.pagename = b'captcha.html'
        self.root = Resource()
        # (None, None) is the (distributor, scheduleInterval):
        self.protectedResource = HTTPServer.WebResourceBridges(None, None)
        self.captchaResource = HTTPServer.GimpCaptchaProtectedResource(
            secretKey='42',
            publicKey='23',
            hmacKey='abcdefghijklmnopqrstuvwxyz012345',
            captchaDir='captchas',
            useForwardedHeader=True,
            protectedResource=self.protectedResource)

        self.root.putChild(self.pagename, self.captchaResource)

        # Set up the basic parts of our faked request:
        self.request = DummyRequest([self.pagename])

    def tearDown(self):
        """Delete the cached CAPTCHA directory if it still exists."""
        if os.path.isdir(self.captchaDir):
            shutil.rmtree(self.captchaDir)

    def test_extractClientSolution(self):
        """A (challenge, sollution) pair extracted from a request resulting
        from a POST should have the same unmodified (challenge, sollution) as
        the client originally POSTed.
        """
        expectedChallenge = '23232323232323232323'
        expectedResponse = 'awefawefaefawefaewf'

        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', expectedChallenge)
        self.request.addArg('captcha_response_field', expectedResponse)

        response = self.captchaResource.extractClientSolution(self.request)
        (challenge, response) = response
        self.assertEqual(challenge, expectedChallenge)
        self.assertEqual(response, expectedResponse)

    def test_checkSolution(self):
        """checkSolution() should return False is the solution is invalid."""
        expectedChallenge = '23232323232323232323'
        expectedResponse = 'awefawefaefawefaewf'

        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', expectedChallenge)
        self.request.addArg('captcha_response_field', expectedResponse)

        valid = self.captchaResource.checkSolution(self.request)
        self.assertFalse(valid)

    def test_getCaptchaImage(self):
        """Retrieving a (captcha, challenge) pair with an empty captchaDir
        should return None for both of the (captcha, challenge) strings.
        """
        self.request.method = b'GET'
        response = self.captchaResource.getCaptchaImage(self.request)
        (image, challenge) = response
        # Because we created the directory, there weren't any CAPTCHAs to
        # retrieve from it:
        self.assertIs(image, None)
        self.assertIs(challenge, None)

    def test_getCaptchaImage_noCaptchaDir(self):
        """Retrieving a (captcha, challenge) with an missing captchaDir should
        raise a bridgedb.captcha.GimpCaptchaError.
        """
        shutil.rmtree(self.captchaDir)
        self.request.method = b'GET'
        self.assertRaises(HTTPServer.captcha.GimpCaptchaError,
                          self.captchaResource.getCaptchaImage, self.request)

    def test_render_GET_missingTemplate(self):
        """render_GET() with a missing template should raise an error and
        return the result of replaceErrorPage().
        """
        oldLookup = HTTPServer.lookup
        try:
            HTTPServer.lookup = None
            self.request.method = b'GET'
            page = self.captchaResource.render_GET(self.request)
            errorPage = HTTPServer.replaceErrorPage(Exception('kablam'))
            self.assertEqual(page, errorPage)
        finally:
            HTTPServer.lookup = oldLookup

    def test_render_POST_blankFields(self):
        """render_POST() with a blank 'captcha_response_field' should return
        a redirect to the CaptchaProtectedResource page.
        """
        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', '')
        self.request.addArg('captcha_response_field', '')

        page = self.captchaResource.render_POST(self.request)
        self.assertEqual(BeautifulSoup(page).find('meta')['http-equiv'],
                         'refresh')

    def test_render_POST_wrongSolution(self):
        """render_POST() with a wrong 'captcha_response_field' should return
        a redirect to the CaptchaProtectedResource page.
        """
        expectedChallenge = '23232323232323232323'
        expectedResponse = 'awefawefaefawefaewf'

        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', expectedChallenge)
        self.request.addArg('captcha_response_field', expectedResponse)

        page = self.captchaResource.render_POST(self.request)
        self.assertEqual(BeautifulSoup(page).find('meta')['http-equiv'],
                         'refresh')


class ReCaptchaProtectedResourceTests(unittest.TestCase):
    """Tests for :mod:`bridgedb.HTTPServer.ReCaptchaProtectedResource`."""

    def setUp(self):
        """Create a :class:`HTTPServer.WebResourceBridges` and protect it with
        a :class:`ReCaptchaProtectedResource`.
        """
        self.timeout = 10.0  # Can't take longer than that, right?
        # Set up our resources to fake a minimal HTTP(S) server:
        self.pagename = b'captcha.html'
        self.root = Resource()
        # (None, None) is the (distributor, scheduleInterval):
        self.protectedResource = HTTPServer.WebResourceBridges(None, None)
        self.captchaResource = HTTPServer.ReCaptchaProtectedResource(
            publicKey='23',
            secretKey='42',
            remoteIP='111.111.111.111',
            useForwardedHeader=True,
            protectedResource=self.protectedResource)

        self.root.putChild(self.pagename, self.captchaResource)

        # Set up the basic parts of our faked request:
        self.request = DummyRequest([self.pagename])

    def tearDown(self):
        """Cleanup method for removing timed out connections on the reactor.

        This seems to be the solution for the dirty reactor due to
        ``DelayedCall``s which is mentioned at the beginning of this
        file. There doesn't seem to be any documentation anywhere which
        proposes this solution, although this seems to solve the problem.
        """
        for delay in reactor.getDelayedCalls():
            try:
                delay.cancel()
            except (AlreadyCalled, AlreadyCancelled):
                pass

    def test_renderDeferred_invalid(self):
        """:meth:`_renderDeferred` should redirect a ``Request`` (after the
        CAPTCHA was NOT xsuccessfully solved) which results from a
        ``Deferred``'s callback.
        """
        self.request.method = b'POST'

        def testCB(request):
            """Check the ``Request`` returned from ``_renderDeferred``."""
            self.assertIsInstance(request, DummyRequest)
            soup = BeautifulSoup(b''.join(request.written)).find('meta')['http-equiv']
            self.assertEqual(soup, 'refresh')

        d = task.deferLater(reactor, 0, lambda x: x, (False, self.request))
        d.addCallback(self.captchaResource._renderDeferred)
        d.addCallback(testCB)
        return d

    def test_renderDeferred_valid(self):
        """:meth:`_renderDeferred` should correctly render a ``Request`` (after
        the CAPTCHA has been successfully solved) which results from a
        ``Deferred``'s callback.
        """
        self.request.method = b'POST'

        def testCB(request):
            """Check the ``Request`` returned from ``_renderDeferred``."""
            self.assertIsInstance(request, DummyRequest)
            html = b''.join(request.written)
            self.assertSubstring('Uh oh, spaghettios!', html)

        d = task.deferLater(reactor, 0, lambda x: x, (True, self.request))
        d.addCallback(self.captchaResource._renderDeferred)
        d.addCallback(testCB)
        return d

    def test_renderDeferred_nontuple(self):
        """:meth:`_renderDeferred` should correctly render a ``Request`` (after
        the CAPTCHA has been successfully solved) which results from a
        ``Deferred``'s callback.
        """
        self.request.method = b'POST'

        def testCB(request):
            """Check the ``Request`` returned from ``_renderDeferred``."""
            self.assertIs(request, None)

        d = task.deferLater(reactor, 0, lambda x: x, (self.request))
        d.addCallback(self.captchaResource._renderDeferred)
        d.addCallback(testCB)
        return d

    def test_checkSolution_blankFields(self):
        """:meth:`HTTPServer.ReCaptchaProtectedResource.checkSolution` should
        return a redirect if is the solution field is blank.
        """
        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', '')
        self.request.addArg('captcha_response_field', '')

        self.assertEqual((False, self.request),
                         self.successResultOf(
                             self.captchaResource.checkSolution(self.request)))

    def test_getRemoteIP_useRandomIP(self):
        """Check that removing our remoteip setting produces a random IP."""
        self.captchaResource.remoteIP = None
        ip = self.captchaResource.getRemoteIP()
        realishIP = ipaddr.IPv4Address(ip).compressed
        self.assertTrue(realishIP)
        self.assertNotEquals(realishIP, '111.111.111.111')

    def test_getRemoteIP_useConfiguredIP(self):
        """Check that our remoteip setting is used if configured."""
        ip = self.captchaResource.getRemoteIP()
        realishIP = ipaddr.IPv4Address(ip).compressed
        self.assertTrue(realishIP)
        self.assertEquals(realishIP, '111.111.111.111')

    def test_render_GET_missingTemplate(self):
        """render_GET() with a missing template should raise an error and
        return the result of replaceErrorPage().
        """
        oldLookup = HTTPServer.lookup
        try:
            HTTPServer.lookup = None
            self.request.method = b'GET'
            page = self.captchaResource.render_GET(self.request)
            errorPage = HTTPServer.replaceErrorPage(Exception('kablam'))
            self.assertEqual(page, errorPage)
        finally:
            HTTPServer.lookup = oldLookup

    def test_render_POST_blankFields(self):
        """render_POST() with a blank 'captcha_response_field' should return
        a redirect to the CaptchaProtectedResource page.
        """
        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', '')
        self.request.addArg('captcha_response_field', '')

        page = self.captchaResource.render_POST(self.request)
        self.assertEqual(page, HTTPServer.server.NOT_DONE_YET)

    def test_render_POST_wrongSolution(self):
        """render_POST() with a wrong 'captcha_response_field' should return
        a redirect to the CaptchaProtectedResource page.
        """
        expectedChallenge = '23232323232323232323'
        expectedResponse = 'awefawefaefawefaewf'

        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', expectedChallenge)
        self.request.addArg('captcha_response_field', expectedResponse)

        page = self.captchaResource.render_POST(self.request)
        self.assertEqual(page, HTTPServer.server.NOT_DONE_YET)


class DummyBridge(object):
    """A mock :class:`bridgedb.Bridges.Bridge` which only supports a mocked
    ``getConfigLine`` method."""

    def _randORPort(self): return random.randint(9001, 9999)
    def _randPTPort(self): return random.randint(6001, 6666)
    def _returnFour(self): return random.randint(2**24, 2**32-1)
    def _returnSix(self): return random.randint(2**24, 2**128-1)

    def __init__(self, transports=[]):
        """Create a mocked bridge suitable for testing distributors and web
        resource rendering.
        """
        self.nickname = "bridge-{0}".format(self._returnFour())
        self.ip = ipaddr.IPv4Address(self._returnFour())
        self.orport = self._randORPort()
        self.transports = transports
        self.running = True
        self.stable = True
        self.blockingCountries = {}
        self.desc_digest = None
        self.ei_digest = None
        self.verified = False
        self.fingerprint = "".join(random.choice('abcdef0123456789')
                                   for _ in xrange(40))
        self.or_addresses = {ipaddr.IPv6Address(self._returnSix()):
                             self._randORPort()}

    def getConfigLine(self, includeFingerprint=True,
                      addressClass=ipaddr.IPv4Address,
                      transport=None,
                      request=None):
        """Get a "torrc" bridge config line to give to a client."""
        line = []
        if transport is not None:
            #print("getConfigLine got transport=%r" % transport)
            line.append(str(transport))
        line.append("%s:%s" % (self.ip, self.orport))
        if includeFingerprint is True: line.append(self.fingerprint)
        bridgeLine = " ".join([item for item in line])
        #print "Created config line: %r" % bridgeLine
        return bridgeLine


class DummyIPBasedDistributor(object):
    """A mocked :class:`bridgedb.Dist.IPBasedDistributor` which is used to test
    :class:`bridgedb.HTTPServer.WebResourceBridges.
    """

    def _dumbAreaMapper(ip): return ip

    def __init__(self, areaMapper=None, nClusters=None, key=None,
                 ipCategories=None, answerParameters=None):
        """None of the parameters are really used, they are just there to retain
        an identical method signature.
        """
        self.areaMapper = self._dumbAreaMapper
        self.nClusters = 3
        self.nBridgesToGive = 3
        self.key = self.__class__.__name__
        self.ipCategories = ipCategories
        self.answerParameters = answerParameters

    def getBridgesForIP(self, ip=None, epoch=None, N=1,
                        countyCode=None, bridgeFilterRules=None):
        """Needed because it's called in
        :meth:`WebResourceBridges.getBridgesForIP`.
        """
        return [DummyBridge() for _ in xrange(N)]


class DummyRequest(requesthelper.DummyRequest):
    """Wrapper for :api:`twisted.test.requesthelper.DummyRequest` to add
    redirect support.
    """

    def __init__(self, *args, **kwargs):
        requesthelper.DummyRequest.__init__(self, *args, **kwargs)
        self.redirect = self._redirect(self)

    def URLPath(self):
        """Fake the missing Request.URLPath too."""
        return self.uri

    def _redirect(self, request):
        """Stub method to add a redirect() method to DummyResponse."""
        newRequest = type(request)
        newRequest.uri = request.uri
        return newRequest


class WebResourceBridgesTests(unittest.TestCase):
    """Tests for :class:`HTTPServer.WebResourceBridges`."""

    def setUp(self):
        """Set up our resources to fake a minimal HTTP(S) server."""
        self.pagename = b'bridges.html'
        self.root = Resource()

        self.dist = DummyIPBasedDistributor()
        self.sched = ScheduledInterval('hour', 1)
        self.nBridgesPerRequest = 2
        self.bridgesResource = HTTPServer.WebResourceBridges(
            self.dist, self.sched, N=2,
            #useForwardedHeader=True,
            includeFingerprints=True)

        self.root.putChild(self.pagename, self.bridgesResource)

    def parseBridgesFromHTMLPage(self, page):
        """Utility to pull the bridge lines out of an HTML response page.

        :param str page: A rendered HTML page, as a string.
        :raises: Any error which might occur.
        :rtype: list
        :returns: A list of the bridge lines contained on the **page**.
        """
        # The bridge lines are contained in a <div class='bridges'> tag:
        soup = BeautifulSoup(page)
        well = soup.find('div', {'class': 'bridge-lines'})
        content = well.renderContents().strip()
        lines = content.splitlines()

        bridges = []
        for line in lines:
            bridgelines = line.split('<br />')
            for bridge in bridgelines:
                if bridge:  # It still could be an empty string at this point
                    bridges.append(bridge)

        return bridges

    def test_render_GET_vanilla(self):
        """Test rendering a request for normal, vanilla bridges."""
        request = DummyRequest([self.pagename])
        request.method = b'GET'
        request.getClientIP = lambda: '1.1.1.1'

        page = self.bridgesResource.render(request)

        # The response should explain how to use the bridge lines:
        self.assertTrue("To enter bridges into Tor Browser" in str(page))

        for b in self.parseBridgesFromHTMLPage(page):
            # Check that each bridge line had the expected number of fields:
            fields = b.split(' ')
            self.assertEqual(len(fields), 2)

            # Check that the IP and port seem okay:
            ip, port = fields[0].rsplit(':')
            self.assertIsInstance(ipaddr.IPv4Address(ip), ipaddr.IPv4Address)
            self.assertIsInstance(int(port), int)
            self.assertGreater(int(port), 0)
            self.assertLessEqual(int(port), 65535)

    def test_render_GET_XForwardedFor(self):
        """The client's IP address should be obtainable from the
        'X-Forwarded-For' header in the request.
        """
        self.bridgesResource.useForwardedHeader = True
        request = DummyRequest([self.pagename])
        request.method = b'GET'
        # Since we do not set ``request.getClientIP`` here like we do in some
        # of the other unittests, an exception would be raised here if
        # ``getBridgesForRequest()`` is unable to get the IP address from this
        # 'X-Forwarded-For' header (because ``ip`` would get set to ``None``).
        request.headers.update({'x-forwarded-for': '2.2.2.2'})

        page = self.bridgesResource.render(request)
        self.bridgesResource.useForwardedHeader = False  # Reset it

        # The response should explain how to use the bridge lines:
        self.assertTrue("To enter bridges into Tor Browser" in str(page))

    def test_render_GET_RTLlang(self):
        """Test rendering a request for plain bridges in Arabic."""
        request = DummyRequest([b"bridges?transport=obfs3"])
        request.method = b'GET'
        request.getClientIP = lambda: '3.3.3.3'
        # For some strange reason, the 'Accept-Language' value *should not* be
        # a list, unlike all the other headers and args…
        request.headers.update({'accept-language': 'ar,en,en_US,'})

        page = self.bridgesResource.render(request)
        self.assertSubstring("direction: rtl", page)
        self.assertSubstring(
            # "I need an alternative way to get bridges!"
            "أحتاج إلى وسيلة بديلة للحصول على bridges", page)

        for bridgeLine in self.parseBridgesFromHTMLPage(page):
            # Check that each bridge line had the expected number of fields:
            bridgeLine = bridgeLine.split(' ')
            self.assertEqual(len(bridgeLine), 2)

    def test_render_GET_RTLlang_obfs3(self):
        """Test rendering a request for obfs3 bridges in Farsi."""
        request = DummyRequest([b"bridges?transport=obfs3"])
        request.method = b'GET'
        request.getClientIP = lambda: '3.3.3.3'
        request.headers.update({'accept-language': 'fa,en,en_US,'})
        # We actually have to set the request args manually when using a
        # DummyRequest:
        request.args.update({'transport': ['obfs3']})

        page = self.bridgesResource.render(request)
        self.assertSubstring("direction: rtl", page)
        self.assertSubstring(
            # "How to use the above bridge lines" (since there should be
            # bridges in this response, we don't tell them about alternative
            # mechanisms for getting bridges)
            "چگونگی از پل‌های خود استفاده کنید", page)

        for bridgeLine in self.parseBridgesFromHTMLPage(page):
            # Check that each bridge line had the expected number of fields:
            bridgeLine = bridgeLine.split(' ')
            self.assertEqual(len(bridgeLine), 3)
            self.assertEqual(bridgeLine[0], 'obfs3')

            # Check that the IP and port seem okay:
            ip, port = bridgeLine[1].rsplit(':')
            self.assertIsInstance(ipaddr.IPv4Address(ip), ipaddr.IPv4Address)
            self.assertIsInstance(int(port), int)
            self.assertGreater(int(port), 0)
            self.assertLessEqual(int(port), 65535)


    def test_renderAnswer_textplain(self):
        """If the request format specifies 'plain', we should return content
        with mimetype 'text/plain'.
        """
        request = DummyRequest([self.pagename])
        request.args.update({'format': ['plain']})
        request.getClientIP = lambda: '4.4.4.4'
        request.method = b'GET'

        page = self.bridgesResource.render(request)
        self.assertTrue("html" not in str(page))

        # We just need to strip and split it because it looks like:
        #
        #   94.235.85.233:9492 0d9d0547c3471cddc473f7288a6abfb54562dc06
        #   255.225.204.145:9511 1fb89d618b3a12afe3529fd072127ea08fb50466
        #
        # (Yes, there are two leading spaces at the beginning of each line)
        #
        bridgeLines = [line.strip() for line in page.strip().split('\n')]

        for bridgeLine in bridgeLines:
            bridgeLine = bridgeLine.split(' ')
            self.assertEqual(len(bridgeLine), 2)

            # Check that the IP and port seem okay:
            ip, port = bridgeLine[0].rsplit(':')
            self.assertIsInstance(ipaddr.IPv4Address(ip), ipaddr.IPv4Address)
            self.assertIsInstance(int(port), int)
            self.assertGreater(int(port), 0)
            self.assertLessEqual(int(port), 65535)


class WebResourceOptionsTests(unittest.TestCase):
    """Tests for :class:`bridgedb.HTTPServer.WebResourceOptions`."""

    def setUp(self):
        """Create a :class:`HTTPServer.WebResourceOptions`."""
        # Set up our resources to fake a minimal HTTP(S) server:
        self.pagename = b'options.html'
        self.root = Resource()
        self.optionsResource = HTTPServer.WebResourceOptions()
        self.root.putChild(self.pagename, self.optionsResource)

    def test_render_GET_RTLlang(self):
        """Test rendering a request for obfs3 bridges in Arabic."""
        request = DummyRequest(["bridges?transport=obfs3"])
        request.method = b'GET'
        request.getClientIP = lambda: '3.3.3.3'
        request.headers.update({'accept-language': 'he'})
        # We actually have to set the request args manually when using a
        # DummyRequest:
        request.args.update({'transport': ['obfs2']})

        page = self.optionsResource.render(request)
        self.assertSubstring("direction: rtl", page)
        self.assertSubstring("מהם גשרים?", page)



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_bridgerequest ; -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2014-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________


import logging

import ipaddr

from zope.interface import implements
from zope.interface import Attribute
from zope.interface import Interface

from bridgedb import Filters
from bridgedb.crypto import getHMACFunc


class IRequestBridges(Interface):
    """Interface specification of client options for requested bridges."""

    addressClass = Attribute(
        "The IP version of bridge addresses to distribute to the client.")
    filters = Attribute(
        "A list of callables used to filter bridges from a hashring.")
    transports = Attribute(
        "A list of strings of Pluggable Transport types requested.")
    notBlockedIn = Attribute(
        "A list of 2-4 letter country codes. The distributed bridges should "
        "not be blocked in these countries.")
    valid = Attribute(
        "A boolean. Should be ``True`` if the client's request was valid.")
    client = Attribute(
        "This should be some information unique to the client making the "
        "request for bridges, such that we are able to HMAC this unique "
        "data, via getHashringPlacement(), in order to place the client "
        "into a hashring (determining which bridge addresses they get in "
        "the request response).")

    def addFilter():
        """Add a filter to the list of ``filters``."""

    def clearFilters():
        """Clear the list of ``filters``."""

    def generateFilters():
        """Build the list of callables, ``filters``, according to the current
        contents of the lists of ``transports``, ``notBlockedIn``, and the
        ``addressClass``.
        """

    def getHashringPlacement():
        """Use some unique parameters of the client making this request to
        obtain a value which we can use to place them into one of the hashrings
        with :class:`~bridgedb.bridges.Bridge`s in it, in order to give that
        client different bridges than other clients.
        """

    def isValid():
        """Determine if the request is ``valid`` according to some parameters."""

    def withIPv4():
        """Set the ``addressClass`` to IPv4."""

    def withIPv6():
        """Set the ``addressClass`` to IPv6."""

    def withPluggableTransportType(typeOfPT):
        """Add this **typeOfPT** to the list of requested ``transports``."""

    def withoutBlockInCountry(countryCode):
        """Add this **countryCode** to the list of countries which distributed
        bridges should not be blocked in (``notBlockedIn``).
        """


class BridgeRequestBase(object):
    """A generic base class for storing options of a client bridge request."""

    implements(IRequestBridges)

    def __init__(self, addressClass=None):
        self.addressClass = addressClass
        if not ((self.addressClass is ipaddr.IPv4Address) or
                (self.addressClass is ipaddr.IPv6Address)):
            self.addressClass = ipaddr.IPv4Address
        self.filters = list()
        self.transports = list()
        self.notBlockedIn = list()
        #: This should be some information unique to the client making the
        #: request for bridges, such that we are able to HMAC this unique data
        #: in order to place the client into a hashring (determining which
        #: bridge addresses they get in the request response). It defaults to
        #: the string ``'default'``.
        self.client = 'default'
        self.valid = False

    def getHashringPlacement(self, key, client=None):
        """Create an HMAC of some **client** info using a **key**.

        :param str key: The key to use for HMACing.
        :param str client: Some (hopefully unique) information about the
            client who is requesting bridges, such as an IP or email address.
        :rtype: long
        :returns: A long specifying index of the first node in a hashring to
            be distributed to the client. This value should obviously be used
            mod the number of nodes in the hashring.
        """
        if not client:
            client = self.client

        # Get an HMAC with the key of the client identifier:
        digest = getHMACFunc(key)(client)
        # Take the lower 8 bytes of the digest and convert to a long:
        position = long(digest[:8], 16)
        return position

    def isValid(self, valid=None):
        """Set or determine if this request was valid.

        :type valid: None or bool
        :param valid: If ``None``, get the current request validity. If
            ``True`` or ``False``, set the request validity accordingly.
        :rtype: bool
        :returns: Whether or not this request is valid.
        """
        if isinstance(valid, bool):
            self.valid = valid
        return self.valid

    def withIPv4(self):
        self.addressClass = ipaddr.IPv4Address

    def withIPv6(self):
        self.addressClass = ipaddr.IPv6Address

    def withoutBlockInCountry(self, country):
        self.notBlockedIn.append(country)

    def withPluggableTransportType(self, pt):
        self.transports.append(pt)

    def addFilter(self, filtre):
        self.filters.append(filtre)

    def clearFilters(self):
        self.filters = []

    def justOnePTType(self):
        """Get just one bridge PT type at a time!"""
        ptType = None
        try:
            ptType = self.transports[-1]  # Use the last PT requested
        except IndexError:
            logging.debug("No pluggable transports were requested.")
        return ptType

    def generateFilters(self):
        if self.addressClass is ipaddr.IPv6Address:
            self.addFilter(Filters.filterBridgesByIP6)
        else:
            self.addFilter(Filters.filterBridgesByIP4)

        transport = self.justOnePTType()
        if transport:
            self.clearFilters()
            self.addFilter(Filters.filterBridgesByTransport(transport,
                                                            self.addressClass))
        for country in self.notBlockedIn:
            self.addFilter(Filters.filterBridgesByNotBlockedIn(country,
                                                               self.addressClass,
                                                               transport))



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2013, Isis Lovecruft
#             (c) 2007-2013, The Tor Project, Inc.
#             (c) 2007-2013, all entities within the AUTHORS file
# :license: 3-Clause BSD, see LICENSE for licensing information

"""Unittests for the :mod:`bridgedb.email.request` module."""

from __future__ import print_function

import ipaddr

from twisted.trial import unittest

from bridgedb.email import request


class DetermineBridgeRequestOptionsTests(unittest.TestCase):
    """Unittests for :func:`b.e.request.determineBridgeRequestOptions`."""

    def test_determineBridgeRequestOptions_get_help(self):
        """Requesting 'get help' should raise EmailRequestedHelp."""
        lines = ['',
                 'get help']
        self.assertRaises(request.EmailRequestedHelp,
                          request.determineBridgeRequestOptions, lines)
        
    def test_determineBridgeRequestOptions_get_halp(self):
        """Requesting 'get halp' should raise EmailRequestedHelp."""
        lines = ['',
                 'get halp']
        self.assertRaises(request.EmailRequestedHelp,
                          request.determineBridgeRequestOptions, lines)
        
    def test_determineBridgeRequestOptions_get_key(self):
        """Requesting 'get key' should raise EmailRequestedKey."""
        lines = ['',
                 'get key']
        self.assertRaises(request.EmailRequestedKey,
                          request.determineBridgeRequestOptions, lines)

    def test_determineBridgeRequestOptions_multiline_invalid(self):
        """Requests without a 'get' anywhere should be considered invalid."""
        lines = ['',
                 'transport obfs3',
                 'ipv6 vanilla bridges',
                 'give me your gpgs']
        reqvest = request.determineBridgeRequestOptions(lines)
        # It's invalid because it didn't include a 'get' anywhere.
        self.assertEqual(reqvest.isValid(), False)
        self.assertFalse(reqvest.wantsKey())
        # Though they did request IPv6, technically.
        self.assertIs(reqvest.addressClass, ipaddr.IPv6Address)
        # And they did request a transport, technically.
        self.assertEqual(len(reqvest.transports), 1)
        self.assertEqual(reqvest.transports[0], 'obfs3')

    def test_determineBridgeRequestOptions_multiline_valid(self):
        """Though requests with a 'get' are considered valid."""
        lines = ['',
                 'get transport obfs3',
                 'vanilla bridges',
                 'transport scramblesuit unblocked ca']
        reqvest = request.determineBridgeRequestOptions(lines)
        # It's valid because it included a 'get'.
        self.assertEqual(reqvest.isValid(), True)
        self.assertFalse(reqvest.wantsKey())
        # Though they didn't request IPv6, so it should default to IPv4.
        self.assertIs(reqvest.addressClass, ipaddr.IPv4Address)
        # And they requested two transports.
        self.assertEqual(len(reqvest.transports), 2)
        self.assertEqual(reqvest.transports[0], 'obfs3')
        self.assertEqual(reqvest.transports[1], 'scramblesuit')
        # And they wanted this stuff to not be blocked in Canada.
        self.assertEqual(len(reqvest.notBlockedIn), 1)
        self.assertEqual(reqvest.notBlockedIn[0], 'ca')

    def test_determineBridgeRequestOptions_multiline_valid_OMG_CAPSLOCK(self):
        """Though requests with a 'get' are considered valid, even if they
        appear to not know the difference between Capslock and Shift.
        """
        lines = ['',
                 'get TRANSPORT obfs3',
                 'vanilla bridges',
                 'TRANSPORT SCRAMBLESUIT UNBLOCKED CA']
        reqvest = request.determineBridgeRequestOptions(lines)
        # It's valid because it included a 'get'.
        self.assertEqual(reqvest.isValid(), True)
        self.assertFalse(reqvest.wantsKey())
        # Though they didn't request IPv6, so it should default to IPv4.
        self.assertIs(reqvest.addressClass, ipaddr.IPv4Address)
        # And they requested two transports.
        self.assertEqual(len(reqvest.transports), 2)
        self.assertEqual(reqvest.transports[0], 'obfs3')
        self.assertEqual(reqvest.transports[1], 'scramblesuit')
        # And they wanted this stuff to not be blocked in Canada.
        self.assertEqual(len(reqvest.notBlockedIn), 1)
        self.assertEqual(reqvest.notBlockedIn[0], 'ca')

    def test_determineBridgeRequestOptions_get_transport(self):
        """An invalid request for 'transport obfs3' (missing the 'get')."""
        lines = ['',
                 'transport obfs3']
        reqvest = request.determineBridgeRequestOptions(lines)
        self.assertEqual(len(reqvest.transports), 1)
        self.assertEqual(reqvest.transports[0], 'obfs3')
        self.assertEqual(reqvest.isValid(), False)
        
    def test_determineBridgeRequestOptions_get_ipv6(self):
        """An valid request for 'get ipv6'."""
        lines = ['',
                 'get ipv6']
        reqvest = request.determineBridgeRequestOptions(lines)
        self.assertIs(reqvest.addressClass, ipaddr.IPv6Address)
        self.assertEqual(reqvest.isValid(), True)


class EmailBridgeRequestTests(unittest.TestCase):
    """Unittests for :class:`b.e.request.EmailBridgeRequest`."""

    def setUp(self):
        """Create an EmailBridgeRequest instance to test."""
        self.request = request.EmailBridgeRequest()

    def tearDown(self):
        """Reset cached 'unblocked'/'transport' lists and addressClass between
        tests.
        """
        self.request.withIPv4()
        self.request.notBlockedIn = []
        self.request.transports = []

    def test_EmailBridgeRequest_isValid_initial(self):
        """Initial value of EmailBridgeRequest.isValid() should be False."""
        self.request.isValid(None)
        self.assertEqual(self.request.isValid(), False)

    def test_EmailBridgeRequest_isValid_True(self):
        """The value of EmailBridgeRequest.isValid() should be True, after it
        has been called with ``True`` as an argument.
        """
        self.request.isValid(True)
        self.assertEqual(self.request.isValid(), True)

    def test_EmailBridgeRequest_isValid_False(self):
        """The value of EmailBridgeRequest.isValid() should be False, after it
        has been called with ``False`` as an argument.
        """
        self.request.isValid(False)
        self.assertEqual(self.request.isValid(), False)

    def test_EmailBridgeRequest_wantsKey_initial(self):
        """Initial value of EmailBridgeRequest.wantsKey() should be False."""
        self.request.wantsKey(None)
        self.assertEqual(self.request.wantsKey(), False)

    def test_EmailBridgeRequest_wantsKey_True(self):
        """The value of EmailBridgeRequest.wantsKey() should be True, after it
        has been called with ``True`` as an argument.
        """
        self.request.wantsKey(True)
        self.assertEqual(self.request.wantsKey(), True)

    def test_EmailBridgeRequest_wantsKey_False(self):
        """The value of EmailBridgeRequest.wantsKey() should be False, after
        it has been called with ``False`` as an argument.
        """
        self.request.wantsKey(False)
        self.assertEqual(self.request.wantsKey(), False)

    def test_EmailBridgeRequest_withIPv6(self):
        """IPv6 requests should have ``addressClass = ipaddr.IPv6Address``."""
        self.assertEqual(self.request.addressClass, ipaddr.IPv4Address)
        self.request.withIPv6()
        self.assertEqual(self.request.addressClass, ipaddr.IPv6Address)

    def test_EmailBridgeRequest_withoutBlockInCountry_CN(self):
        """Country codes that aren't lowercase should be ignored."""
        self.request.withoutBlockInCountry('get unblocked CN')
        self.assertIsInstance(self.request.notBlockedIn, list)
        self.assertEqual(len(self.request.notBlockedIn), 0)

    def test_EmailBridgeRequest_withoutBlockInCountry_cn(self):
        """Lowercased country codes are okay though."""
        self.request.withoutBlockInCountry('get unblocked cn')
        self.assertIsInstance(self.request.notBlockedIn, list)
        self.assertEqual(len(self.request.notBlockedIn), 1)

    def test_EmailBridgeRequest_withoutBlockInCountry_cn_getMissing(self):
        """Lowercased country codes are still okay if the 'get' is missing."""
        self.request.withoutBlockInCountry('unblocked cn')
        self.assertIsInstance(self.request.notBlockedIn, list)
        self.assertEqual(len(self.request.notBlockedIn), 1)

    def test_EmailBridgeRequest_withoutBlockInCountry_multiline_cn_ir_li(self):
        """Requests for multiple unblocked countries should compound if they
        are on separate 'get unblocked' lines.
        """
        self.request.withoutBlockInCountry('get unblocked cn')
        self.request.withoutBlockInCountry('get unblocked ir')
        self.request.withoutBlockInCountry('get unblocked li')
        self.assertIsInstance(self.request.notBlockedIn, list)
        self.assertEqual(len(self.request.notBlockedIn), 3)

    def test_EmailBridgeRequest_withoutBlockInCountry_singleline_cn_ir_li(self):
        """Requests for multiple unblocked countries which are all on the same
        'get unblocked' line will use only the *first* country code.
        """
        self.request.withoutBlockInCountry('get unblocked cn ir li')
        self.assertIsInstance(self.request.notBlockedIn, list)
        self.assertEqual(len(self.request.notBlockedIn), 1)

    def test_EmailBridgeRequest_withPluggableTransportType_SCRAMBLESUIT(self):
        """Transports which aren't in lowercase should be ignored."""
        self.request.withPluggableTransportType('get transport SCRAMBLESUIT')
        self.assertIsInstance(self.request.transports, list)
        self.assertEqual(len(self.request.transports), 0)

    def test_EmailBridgeRequest_withPluggableTransportType_scramblesuit(self):
        """Lowercased transports are okay though."""
        self.request.withPluggableTransportType('get transport scramblesuit')
        self.assertIsInstance(self.request.transports, list)
        self.assertEqual(len(self.request.transports), 1)
        self.assertEqual(self.request.transports[0], 'scramblesuit')

    def test_EmailBridgeRequest_withPluggableTransportType_scramblesuit_getMissing(self):
        """Lowercased transports are still okay if 'get' is missing."""
        self.request.withPluggableTransportType('transport scramblesuit')
        self.assertIsInstance(self.request.transports, list)
        self.assertEqual(len(self.request.transports), 1)
        self.assertEqual(self.request.transports[0], 'scramblesuit')

    def test_EmailBridgeRequest_withPluggableTransportType_multiline_obfs3_obfs2_scramblesuit(self):
        """Requests for multiple pluggable transports should compound if they
        are on separate 'get transport' lines.
        """
        self.request.withPluggableTransportType('get transport obfs3')
        self.request.withPluggableTransportType('get transport obfs2')
        self.request.withPluggableTransportType('get transport scramblesuit')
        self.assertIsInstance(self.request.transports, list)
        self.assertEqual(len(self.request.transports), 3)
        self.assertEqual(self.request.transports[0], 'obfs3')

    def test_EmailBridgeRequest_withPluggableTransportType_singleline_obfs3_obfs2_scramblesuit(self):
        """Requests for multiple transports which are all on the same
        'get transport' line will use only the *first* transport.
        """
        self.request.withPluggableTransportType('get transport obfs3 obfs2 scramblesuit')
        self.assertIsInstance(self.request.transports, list)
        self.assertEqual(len(self.request.transports), 1)
        self.assertEqual(self.request.transports[0], 'obfs3')

    def test_EmailBridgeRequest_withPluggableTransportType_whack(self):
        """Requests for whacky transports that don't exist are also okay."""
        self.request.withPluggableTransportType('get transport whack')
        self.assertIsInstance(self.request.transports, list)
        self.assertEqual(len(self.request.transports), 1)
        self.assertEqual(self.request.transports[0], 'whack')

    def test_EmailBridgeRequest_justOnePTType_obfs3_obfs2_scramblesuit(self):
        """Requests for multiple transports when
        ``EmailBridgeRequest.justOneTransport()`` is used will use only the
        *last* transport.
        """
        self.request.withPluggableTransportType('get transport obfs3')
        self.request.withPluggableTransportType('get transport obfs2')
        self.request.withPluggableTransportType('get transport scramblesuit')
        self.assertIsInstance(self.request.transports, list)
        self.assertEqual(len(self.request.transports), 3)
        self.assertEqual(self.request.transports[0], 'obfs3')
        self.assertEqual(self.request.justOnePTType(), 'scramblesuit')



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2013, Isis Lovecruft
#             (c) 2007-2013, The Tor Project, Inc.
#             (c) 2007-2013, all entities within the AUTHORS file
# :license: 3-Clause BSD, see LICENSE for licensing information

"""Class wrappers to adapt BridgeDB old unittests in :mod:`bridgedb.Tests`
(now kept in :mod:`bridgedb.test.legacy_Tests`) to be compatible with the
newer :api:`twisted.trial` unittests in this directory.
"""

from __future__ import print_function
from __future__ import unicode_literals

import binascii
import doctest
import glob
import logging
import os
import warnings

from twisted.python import monkey
from twisted.trial import unittest

from bridgedb.test import legacy_Tests as Tests
from bridgedb.test import deprecated


warnings.filterwarnings('ignore', module="bridgedb\.test\.legacy_Tests")
pyunit = __import__('unittest')


def generateTrialAdaptedDoctestsSuite():
    """Dynamically generates a :class:`unittest.TestSuite` all containing
    discovered doctests within the installed ``bridgedb`` package.
    """
    bridgedb     = __import__('bridgedb')
    fakedGlobals = globals().update({'bridgedb': bridgedb})
    modulePath   = bridgedb.__path__[0]

    #: The package directories to search for source code with doctests.
    packagePaths = [modulePath,
                    os.path.join(modulePath, 'parse'),
                    os.path.join(modulePath, 'test')]
    #: The source code files which will be searched for doctests.
    files = []
    #: The cls.testSuites which the test methods will be generated from.
    testSuites = []

    packages = [os.path.join(pkg, '*.py') for pkg in packagePaths]
    [files.extend(glob.glob(pkg)) for pkg in packages]

    for filename in files:
        testSuites.append(
            doctest.DocFileSuite(filename,
                                 module_relative=False,
                                 globs=fakedGlobals))
    return testSuites

def monkeypatchTests():
    """Monkeypatch the old unittests, replacing new, refactored code with their
    original equivalents from :mod:`bridgedb.test.deprecated`.

    The first patch replaces the newer parsing function,
    :func:`~bridgedb.parse.networkstatus.parseALine`, with the older,
    :func:`deprecated one <bridgedb.test.deprecated.parseORAddressLine>` (the
    old function was previously located at
    ``bridgedb.Bridges.parseORAddressLine``).

    The second patch replaces the new :class:`~bridgedb.parse.addr.PortList`,
    with the :class:`older one <bridgedb.test.deprecated.PortList>` (which
    was previously located at ``bridgedb.Bridges.PortList``).

    The third, forth, and fifth monkeypatches add some module-level attributes
    back into :mod:`bridgedb.Bridges`.

    :rtype: :api:`~twisted.python.monkey.MonkeyPatcher`
    :returns: A :api:`~twisted.python.monkey.MonkeyPatcher`, preloaded with
              patches from :mod:`bridgedb.test.deprecated`.
    """
    patcher = monkey.MonkeyPatcher()
    patcher.addPatch(Tests.networkstatus, 'parseALine',
                     deprecated.parseORAddressLine)
    patcher.addPatch(Tests.addr, 'PortList', deprecated.PortList)
    patcher.addPatch(Tests.bridgedb.Bridges, 'PORTSPEC_LEN', 16)
    patcher.addPatch(Tests.bridgedb.Bridges, 're_ipv4', deprecated.re_ipv4)
    patcher.addPatch(Tests.bridgedb.Bridges, 're_ipv6', deprecated.re_ipv6)
    patcher.addPatch(Tests.bridgedb.Bridges, 'HEX_FP_LEN', 40)
    patcher.addPatch(Tests.bridgedb.Bridges, 'toHex', binascii.b2a_hex)
    patcher.addPatch(Tests.bridgedb.Bridges, 'fromHex', binascii.a2b_hex)
    patcher.addPatch(Tests.bridgedb.Bridges, 'is_valid_fingerprint',
                     deprecated.is_valid_fingerprint)
    patcher.addPatch(Tests.bridgedb.Bridges, 'PluggableTransport',
                     deprecated.PluggableTransport)
    patcher.addPatch(Tests.bridgedb.Bridges, 'Bridge',
                     deprecated.Bridge)
    return patcher


class DynamicTestCaseMeta(type):
    """You know how scary the seemingly-arbitrary constants in elliptic curve
    cryptography seem? Well, I am over nine thousand times more scary. Dynamic
    warez… beware! Be afraid; be very afraid.

    :ivar testResult: An :class:`unittest.TestResult` adapted with
                      :api:`twisted.trial.unittest.PyUnitResultAdapter`, for
                      storing test failures and successes in.

    A base class which uses this metaclass should define the following class
    attributes:

    :ivar testSuites: A list of :class:`unittest.TestSuite`s (or their
                      :mod:`doctest` or :api:`twisted.trial` equivalents).
    :ivar methodPrefix: A string to prefix the generated method names
                        with. (default: 'test_')
    """

    testResult = unittest.PyUnitResultAdapter(pyunit.TestResult())

    def __new__(cls, name, bases, attrs):
        """Construct the initialiser for a new
        :api:`twisted.trial.unittest.TestCase`.
        """
        logging.debug("Metaclass __new__ constructor called for %r" % name)

        if not 'testSuites' in attrs:
            attrs['testSuites'] = list()
        if not 'methodPrefix' in attrs:
            attrs['methodPrefix'] = 'test_'

        testSuites   = attrs['testSuites']
        methodPrefix = attrs['methodPrefix']
        logging.debug(
            "Metaclass __new__() class %r(testSuites=%r, methodPrefix=%r)" %
            (name, '\n\t'.join([str(ts) for ts in testSuites]), methodPrefix))

        generatedMethods = cls.generateTestMethods(testSuites, methodPrefix)
        attrs.update(generatedMethods)
        #attrs['init'] = cls.__init__  # call the standard initialiser
        return super(DynamicTestCaseMeta, cls).__new__(cls, name, bases, attrs)

    @classmethod
    def generateTestMethods(cls, testSuites, methodPrefix='test_'):
        """Dynamically generate methods and their names for a
        :api:`twisted.trial.unittest.TestCase`.

        :param list testSuites: A list of :class:`unittest.TestSuite`s (or
                                their :mod:`doctest` or :api:`twisted.trial`
                                equivalents).
        :param str methodPrefix: A string to prefix the generated method names
                                 with. (default: 'test_')
        :rtype: dict
        :returns: A dictionary of class attributes whose keys are dynamically
                  generated method names (prefixed with **methodPrefix**), and
                  whose corresponding values are dynamically generated methods
                  (taken out of the class attribute ``testSuites``).
        """
        def testMethodFactory(test, name):
            def createTestMethod(test):
                def testMethod(*args, **kwargs):
                    """When this function is generated, a methodname (beginning
                    with whatever **methodPrefix** was set to) will also be
                    generated, and the (methodname, method) pair will be
                    assigned as attributes of the generated
                    :api:`~twisted.trial.unittest.TestCase`.
                    """
                    # Get the number of failures before test.run():
                    origFails = len(cls.testResult.original.failures)
                    test.run(cls.testResult)
                    # Fail the generated testMethod if the underlying failure
                    # count has increased:
                    if (len(cls.testResult.original.failures) > origFails):
                        fail = cls.testResult.original.failures[origFails:][0]
                        raise unittest.FailTest(''.join([str(fail[0]),
                                                         str(fail[1])]))
                    return cls.testResult
                testMethod.__name__ = str(name)
                return testMethod
            return createTestMethod(test)

        newAttrs = {}
        for testSuite in testSuites:
            for test in testSuite:
                origName = test.id()
                if origName.find('.') > 0:
                    origFunc = origName.split('.')[-2:]
                    origName = '_'.join(origFunc)
                if origName.endswith('_py'):  # this happens with doctests
                    origName = origName.strip('_py')
                methName = str(methodPrefix + origName).replace('.', '_')
                meth = testMethodFactory(test, methName)
                logging.debug("Set %s.%s=%r" % (cls.__name__, methName, meth))
                newAttrs[methName] = meth
        return newAttrs


class OldUnittests(unittest.TestCase):
    """A wrapper around :mod:`bridgedb.Tests` to produce :api:`~twisted.trial`
    compatible output.

    Generates a :api:`twisted.trial.unittest.TestCase` containing a
    test for each of the individual tests in :mod:`bridgedb.Tests`.

    Each test in this :api:`~twisted.trial.unittest.TestCase`` is dynamically
    generated from one of the old unittests in :mod:`bridgedb.Tests`. Then,
    the class is wrapped to cause the results reporting mechanisms to be
    :api:`~twisted.trial` compatible.

    :returns: A :api:`twisted.trial.unittest.TestCase`.
    """
    __metaclass__ = DynamicTestCaseMeta
    testSuites    = Tests.testSuite()
    testResult    = unittest.PyUnitResultAdapter(pyunit.TestResult())
    methodPrefix  = 'test_regressionsNewCode_'


class MonkeypatchedOldUnittests(unittest.TestCase):
    """A wrapper around :mod:`bridgedb.Tests` to produce :api:`~twisted.trial`
    compatible output.

    For each test in this ``TestCase``, one of the old unittests in
    bridgedb/Tests.py is run. For all of the tests, some functions and classes
    are :api:`twisted.python.monkey.MonkeyPatcher.patch`ed with old,
    deprecated code from :mod:`bridgedb.test.deprecated` to ensure that any
    new code has not caused any regressions.
    """
    __metaclass__ = DynamicTestCaseMeta
    testSuites    = Tests.testSuite()
    testResult    = unittest.PyUnitResultAdapter(pyunit.TestResult())
    methodPrefix  = 'test_regressionsOldCode_'
    patcher       = monkeypatchTests()

    def runWithPatches(self, *args):
        """Replaces :api:`~twisted.trial.unittest.TestCase.run` as the default
        methodName to run. This method calls ``run()`` though
        ``self.patcher.runWithPatches``, using the class **testResult** object.
        """
        self.patcher.runWithPatches(self.run, self.testResult)
        self.patcher.restore()

    def __init__(self, methodName='runWithPatches'):
        super(MonkeypatchedOldUnittests, self).__init__(methodName=methodName)


class TrialAdaptedDoctests(unittest.TestCase):
    """Discovers and runs all doctests within the ``bridgedb`` package.

    Finds all doctests from the directory that BridgeDB was installed in, in
    all Python modules and packages, and runs them with :api:`twisted.trial`.
    """
    __metaclass__ = DynamicTestCaseMeta
    testSuites    = generateTrialAdaptedDoctestsSuite()
    testResult    = unittest.PyUnitResultAdapter(pyunit.TestResult())
    methodPrefix  = 'test_doctestsIn_'



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2013-2015 Isis Lovecruft
#             (c) 2007-2015, The Tor Project, Inc.
#             (c) 2007-2015, all entities within the AUTHORS file
# :license: 3-clause BSD, see included LICENSE for information

"""Classes for finding and managing lists of open proxies."""

from __future__ import print_function
from collections import MutableSet
from functools import update_wrapper
from functools import wraps

import ipaddr
import logging
import os
import time

from twisted.internet import defer
from twisted.internet import protocol
from twisted.internet import reactor
from twisted.internet import utils as txutils
from bridgedb.runner import find
from bridgedb.parse.addr import isIPAddress


def downloadTorExits(proxyList, ipaddress, port=443, protocol=None):
    """Run a script which downloads a list of Tor exit relays which allow their
    clients to exit to the given **ipaddress** and **port**.

    :param proxyList: The :class:`ProxySet` instance from :mod:`bridgedb.Main`.
    :param str ipaddress: The IP address that each Tor exit relay should be
        capable of connecting to for clients, as specified by its ExitPolicy.
    :param int port: The port corresponding to the above **ipaddress** that
        each Tor exit relay should allow clients to exit to. (See
        https://check.torproject.org/cgi-bin/TorBulkExitList.py.)
    :type protocol: :api:`twisted.internet.protocol.Protocol`
    :param protocol: A :class:`~bridgedb.proxy.ExitListProtocol`, or any other
        :api:`~twisted.internet.protocol.Protocol` implementation for
        processing the results of a process which downloads a list of Tor exit
        relays. This parameter is mainly meant for use in testing, and should
        not be changed.
    :rtype: :class:`~twisted.internet.defer.Deferred`
    :returns: A deferred which will callback with a list, each item in the
        list is a string containing an IP of a Tor exit relay.
    """
    proto = ExitListProtocol() if protocol is None else protocol()
    args = [proto.script, '--stdout', '-a', ipaddress, '-p', str(port)]
    proto.deferred.addCallback(proxyList.addExitRelays)
    proto.deferred.addErrback(logging.exception)
    transport = reactor.spawnProcess(proto, proto.script, args=args, env={})
    return proto.deferred

def loadProxiesFromFile(filename, proxySet=None, removeStale=False):
    """Load proxy IP addresses from a list of files.

    :param str filename: A filename whose path can be either absolute or
        relative to the current working directory. The file should contain the
        IP addresses of various open proxies, one per line, for example::

            11.11.11.11
            22.22.22.22
            123.45.67.89

    :type proxySet: None or :class:`~bridgedb.proxy.ProxySet`.
    :param proxySet: If given, load the addresses read from the files into
        this ``ProxySet``.
    :param bool removeStale: If ``True``, remove proxies from the **proxySet**
        which were not listed in any of the **files**.
        (default: ``False``)
    :returns: A list of all the proxies listed in the **files* (regardless of
        whether they were added or removed).
    """
    logging.info("Reloading proxy lists...")

    addresses = []
    if proxySet:
        oldProxySet = proxySet.copy()

    try:
        with open(filename, 'r') as proxyFile:
            for line in proxyFile.readlines():
                line = line.strip()
                if proxySet:
                    # ProxySet.add() will validate the IP address
                    if proxySet.add(line, tag=filename):
                        logging.info("Added %s to the proxy list." % line)
                        addresses.append(line)
                else:
                    ip = isIPAddress(line)
                    if ip:
                        addresses.append(ip)
    except Exception as error:
        logging.warn("Error while reading a proxy list file: %s" % str(error))

    if proxySet:
        stale = list(oldProxySet.difference(addresses))

        if removeStale:
            for ip in stale:
                if proxySet.getTag(ip) == filename:
                    logging.info("Removing stale IP %s from proxy list." % ip)
                    proxySet.remove(ip)
                else:
                    logging.info("Tag %s didn't match %s"
                                 % (proxySet.getTag(ip), filename))

    return addresses


class ProxySet(MutableSet):
    """A :class:`collections.MutableSet` for storing validated IP addresses."""

    #: A tag to apply to IP addresses within this ``ProxySet`` which are known
    #: Tor exit relays.
    _exitTag = 'exit_relay'

    def __init__(self, proxies=dict()):
        """Initialise a ``ProxySet``.

        :type proxies: A tuple, list, dict, or set.
        :param proxies: Optionally, initialise with an iterable, ``proxies``.
            For each ``item`` in that iterable, ``item`` must either:
                1. be a string or int representing an IP address, or,
                2. be another iterable, whose first item satisfies #1.
        """
        super(ProxySet, self).__init__()
        self._proxydict = dict()
        self._proxies = set()
        self.addProxies(proxies)

    @property
    def proxies(self):
        """All proxies in this set, regardless of tags."""
        return list(self._proxies)

    @property
    def exitRelays(self):
        """Get all proxies in this ``ProxySet`` tagged as Tor exit relays.

        :rtype: set
        :returns: A set of all known Tor exit relays which are contained
            within this :class:`~bridgedb.proxy.ProxySet`.
        """
        return self.getAllWithTag(self._exitTag)

    def __add__(self, ip=None, tag=None):
        """Add an **ip** to this set, with an optional **tag**.

        This has no effect if the **ip** is already present.  The **ip** is
        only added if it passes the checks in
        :func:`~bridgedb.parse.addr.isIPAddress`.

        :type ip: basestring or int
        :param ip: The IP address to add.
        :param tag: An optional value to link to **ip**. If not given, it will
            be a timestamp (seconds since epoch, as a float) for when **ip**
            was first added to this set.
        :rtype: bool
        :returns: ``True`` if **ip** is in this set; ``False`` otherwise.
        """
        ip = isIPAddress(ip)
        if ip:
            if self._proxies.isdisjoint(set(ip)):
                logging.debug("Adding %s to proxy list %r..." % (ip, self))
                self._proxies.add(ip)
                self._proxydict[ip] = tag if tag else time.time()
                return True
        return False

    def __radd__(self, *args, **kwargs): self.__add__(*args, **kwargs)

    def __contains__(self, ip):
        """x.__contains__(y) <==> y in x.

        :type ip: basestring or int
        :param ip: The IP address to check.
        :rtype: boolean
        :returns: True if ``ip`` is in this set; False otherwise.
        """
        ipset = [isIPAddress(ip),]
        if ipset and len(self._proxies.intersection(ipset)) == len(ipset):
            return True
        return False

    def __sub__(self, ip):
        """Entirely remove **ip** from this set.

        :type ip: basestring or int
        :param ip: The IP address to remove.
        """
        try:
            self._proxydict.pop(ip)
            self._proxies.discard(ip)
        except KeyError:
            pass

    def __rsub__(self, *args, **kwargs): raise NotImplemented

    def _getErrorMessage(self, x=None, y=None):
        """Make an error message describing how this class works."""
        message = """\nParameter 'proxies' must be one of:
    - a {1} of {0}
    - a {2} of {0}
    - a {3}, whose keys are {0} (the values can be anything)
    - a {4} of {1}s, whose first object in each {1} must be a {0}
    - a {4} of {0}
        """.format(type(''), type(()), type([]), type({}), type(set(())))
        end = "You gave: a {0}".format(type(y))
        end += " of {0}".format(type(x))
        return os.linesep.join((message, end))

    def addProxies(self, proxies, tag=None):
        """Add proxies to this set.

        This calls :func:`add` for each item in the iterable **proxies**.
        Each proxy, if added, will be tagged with a current timestamp.

        :type proxies: A tuple, list, dict, or set.
        :param proxies: An iterable.  For each ``item`` in that iterable,
            ``item`` must either:
                1. be a string or int representing an IP address, or,
                2. be another iterable, whose first item satisfies #1.
        :keyword tag: An optional value to link to all untagged
            **proxies**. If ``None``, it will be a timestamp (seconds since
            epoch, as a float) for when the proxy was first added to this set.
        """
        if isinstance(proxies, dict):
            [self.add(ip, value) for (ip, value) in proxies.items()]
        else:
            try:
                for x in proxies:
                    if isinstance(x, (tuple, list, set)):
                        if len(x) == 2:   self.add(x[0], x[1])
                        elif len(x) == 1: self.add(x, tag)
                        else: raise ValueError(self._getErrorMessage(x, proxies))
                    elif isinstance(x, (basestring, int)):
                        self.add(x, tag)
                    else:
                        raise ValueError(self._getErrorMessage(x, proxies))
            except TypeError:
                raise ValueError(self._getErrorMessage(proxies, None))

    @wraps(addProxies)
    def addExitRelays(self, relays):
        logging.info("Loading exit relays into proxy list...")
        [self.add(x, self._exitTag) for x in relays]

    def getTag(self, ip):
        """Get the tag for an **ip** in this ``ProxySet``, if available.

        :type ip: basestring or int
        :param ip: The IP address to obtain the tag for.
        :rtype: ``None`` or basestring or int
        :returns: The tag for that **ip**, iff **ip** exists in this
            ``ProxySet`` and it has a tag.
        """
        return self._proxydict.get(ip)

    def getAllWithTag(self, tag):
        """Get all proxies in this ``ProxySet`` with a given tag.

        :param basestring tag: A tag to search for.
        :rtype: set
        :returns: A set of all proxies which are contained within this
            :class:`~bridgedb.proxy.ProxySet` which are also tagged with
            **tag**.
        """
        return set([key for key, value in filter(lambda x: x[1] == tag,
                                                 self._proxydict.items())])

    def firstSeen(self, ip):
        """Get the timestamp when **ip** was first seen, if available.

        :type ip: basestring or int
        :param ip: The IP address to obtain a timestamp for.
        :rtype: float or None
        :returns: The timestamp (in seconds since epoch) if available.
            Otherwise, returns None.
        """
        when = self.getTag(ip)
        if isinstance(when, float):
            return when

    def isExitRelay(self, ip):
        """Check if ``ip`` is a known Tor exit relay.

        :type ip: basestring or int
        :param ip: The IP address to check.
        :rtype: boolean
        :returns: True if ``ip`` is a known Tor exit relay; False otherwise.
        """
        if self.getTag(ip) == self._exitTag:
            return True
        return False

    def replaceProxyList(self, proxies, tag=None):
        """Clear everything and add all ``proxies``.

        :type proxies: A tuple, list, dict, or set.
        :param proxies: An iterable.  For each ``item`` in that iterable,
            ``item`` must either:
                1. be a string or int representing an IP address, or,
                2. be another iterable, whose first item satisfies #1.
        """
        try:
            self.clear()
            self.addProxies(proxies, tag=tag)
        except Exception as error:
            logging.error(str(error))

    _assigned = ('__name__', '__doc__')

    @wraps(MutableSet._hash)
    def __hash__(self):      return self._hash()
    def __iter__(self):      return self._proxies.__iter__()
    def __len__(self):       return len(self._proxydict.items())
    def __repr__(self):      return type('')(self.proxies)
    def __str__(self):       return os.linesep.join(self.proxies)
    update_wrapper(__iter__, set.__iter__, _assigned)
    update_wrapper(__len__,  len, _assigned)
    update_wrapper(__repr__, repr, _assigned)
    update_wrapper(__str__,  str, _assigned)

    def add(self, ip, tag=None): return self.__add__(ip, tag)
    def copy(self):              return self.__class__(self._proxydict.copy())
    def contains(self, ip):      return self.__contains__(ip)
    def discard(self, ip):       return self.__sub__(ip)
    def remove(self, other):     return self.__sub__(other)
    update_wrapper(add,          __add__)
    update_wrapper(copy,         __init__)
    update_wrapper(contains,     __contains__)
    update_wrapper(discard,      __sub__)
    update_wrapper(remove,       __sub__)

    def difference(self, other):           return self._proxies.difference(other)
    def issubset(self, other):             return self._proxies.issubset(other)
    def issuperset(self, other):           return self._proxies.issuperset(other)
    def intersection(self, other):         return self._proxies.intersection(other)
    def symmetric_difference(self, other): return self._proxies.symmetric_difference(other)
    def union(self, other):                return self._proxies.union(other)
    update_wrapper(difference,             set.difference, _assigned)
    update_wrapper(issubset,               set.issubset, _assigned)
    update_wrapper(issuperset,             set.issuperset, _assigned)
    update_wrapper(intersection,           set.intersection, _assigned)
    update_wrapper(symmetric_difference,   set.symmetric_difference, _assigned)
    update_wrapper(union,                  set.union, _assigned)


class ExitListProtocol(protocol.ProcessProtocol):
    """A :class:`~twisted.internet.protocol.Protocol` for ``get-exit-list``.

    :attr boolean connected: True if our ``transport`` is connected.

    :type transport: An implementer of
        :interface:`twisted.internet.interface.IProcessTransport`.
    :attr transport: If :func:`twisted.internet.reactor.spawnProcess` is
        called with an instance of this class as it's ``protocol``, then
        :func:`~twisted.internet.reactor.spawnProcess` will return this
        ``transport``.
    """

    def __init__(self):
        """Create a protocol for downloading a list of current Tor exit relays.

        :type exitlist: :class:`ProxySet`
        :ivar exitlist: A :class:`~collections.MutableSet` containing the IP
            addresses of known Tor exit relays which can reach our public IP
            address.
        :ivar list data: A list containing a ``bytes`` object for each chuck
            of data received from the ``transport``.
        :ivar deferred: A deferred which will callback with the ``exitlist``
            when the process has ended.

        :param string script: The full pathname of the script to run.
        """
        self.data = []
        self.script = find('get-tor-exits')
        self.exitlist = ProxySet()
        self.deferred = defer.Deferred()

    def childConnectionLost(self, childFD):
        """See :func:`t.i.protocol.ProcessProtocol.childConnectionLost`."""
        protocol.ProcessProtocol.childConnectionLost(self, childFD)

    def connectionMade(self):
        """Called when a connection is made.

        This may be considered the initializer of the protocol, because it is
        called when the connection is completed.  For clients, this is called
        once the connection to the server has been established; for servers,
        this is called after an accept() call stops blocking and a socket has
        been received.  If you need to send any greeting or initial message,
        do it here.
        """
        logging.debug("ExitListProtocol: Connection made with remote server")
        self.transport.closeStdin()

    def errReceived(self, data):
        """Some data was received from stderr."""
        # The get-exit-list script uses twisted.python.log to log to stderr:
        logging.debug(data)  # pragma: no cover

    def outReceived(self, data):
        """Some data was received from stdout."""
        self.data.append(data)

    def outConnectionLost(self):
        """This will be called when stdout is closed."""
        logging.debug("Finished downloading list of Tor exit relays.")
        self.transport.loseConnection()
        self.parseData()

    def parseData(self):
        """Parse all data received so far into our
        :class:`<bridgedb.proxy.ProxySet> exitlist`.
        """
        unparseable = []

        data = ''.join(self.data).split('\n')

        for line in data:
            line.strip()
            if not line: continue
            # If it reached an errorpage, then we grabbed raw HTML that starts
            # with an HTML tag:
            if line.startswith('<'): break
            if line.startswith('#'): continue
            ip = isIPAddress(line)
            if ip:
                logging.info("Discovered Tor exit relay: %s" % ip)
                self.exitlist.add(ip)
            else:
                logging.debug("Got exitlist line that wasn't an IP: %s" % line)
                unparseable.append(line)

        if unparseable:
            logging.warn(("There were unparseable lines in the downloaded "
                          "list of Tor exit relays: %r") % unparseable)

    def processEnded(self, reason):
        """Called when the child process exits and all file descriptors
        associated with it have been closed.

        :type reason: :class:`twisted.python.failure.Failure`
        """
        self.transport.loseConnection()
        if reason.value.exitCode != 0:  # pragma: no cover
            logging.debug(reason.getTraceback())
            logging.error("There was an error downloading Tor exit list: %s"
                          % reason.value)
        else:
            logging.info("Finished processing list of Tor exit relays.")
        logging.debug("Transferring exit list to storage...")
        # Avoid triggering the deferred twice, e.g. on processExited():
        if not self.deferred.called:
            self.deferred.callback(list(self.exitlist.proxies))

    def processExited(self, reason):
        """This will be called when the subprocess exits.

        :type reason: :class:`twisted.python.failure.Failure`
        """
        logging.debug("%s exited with status code %d"
                      % (self.script, reason.value.exitCode))



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2013-2015 Isis Lovecruft
#             (c) 2007-2015, The Tor Project, Inc.
#             (c) 2007-2015, all entities within the AUTHORS file
# :license: 3-clause BSD, see included LICENSE for information

"""Parsers for bridge networkstatus descriptors.

THIS ENTIRE MODULE WAS DEPRECATED (AS PART OF #9380_) AND WAS REPLACED WITH
THE CORRESPONDING FUNCTIONS IN :mod:`bridgedb.parse.descriptors`.

.. #9380: https://bugs.torproject.org/9380

.. py:module:: bridgedb.parse.networkstatus
   :synopsis: Parsers for ``@type bridge-network-status`` descriptors_.
.. _descriptors: https://metrics.torproject.org/formats.html#descriptortypes


bridgedb.parse.networkstatus
============================
::

  networkstatus
   |_ isValidRouterNickname - Determine if a nickname is according to spec
   |_ parseRLine - Parse an 'r'-line from a networkstatus document
   |_ parseALine - Parse an 'a'-line from a networkstatus document
    \_ parseSLine - Parse an 's'-line from a networkstatus document
..
"""

import binascii
import logging
import string
import time
import warnings

from twisted.python import deprecate
from twisted.python.log import showwarning
from twisted.python.versions import Version

from bridgedb.parse import addr
from bridgedb.parse import parseUnpaddedBase64
from bridgedb.parse import InvalidBase64
from bridgedb.parse.nickname import InvalidRouterNickname
from bridgedb.parse.nickname import isValidRouterNickname


@deprecate.deprecated(Version('bridgedb', 0, 2, 5))
class NetworkstatusParsingError(Exception):
    """Unable to parse networkstatus document line."""

@deprecate.deprecated(Version('bridgedb', 0, 2, 5))
class InvalidNetworkstatusRouterIdentity(ValueError):
    """The ID field of a networkstatus document 'r'-line is invalid."""


@deprecate.deprecated(Version('bridgedb', 0, 2, 5))
def parseRLine(line):
    """Parse an 'r'-line from a networkstatus document.

    From torspec.git/dir-spec.txt, commit 36761c7d553d L1499-1512:
      | "r" SP nickname SP identity SP digest SP publication SP IP SP ORPort
      |     SP DirPort NL
      |
      |     [At start, exactly once.]
      |
      |     "Nickname" is the OR's nickname.  "Identity" is a hash of its
      |     identity key, encoded in base64, with trailing equals sign(s)
      |     removed.  "Digest" is a hash of its most recent descriptor as
      |     signed (that is, not including the signature), encoded in base64.
      |     "Publication" is the
      |     publication time of its most recent descriptor, in the form
      |     YYYY-MM-DD HH:MM:SS, in UTC.  "IP" is its current IP address;
      |     ORPort is its current OR port, "DirPort" is its current directory
      |     port, or "0" for "none".

    :param string line: An 'r'-line from an bridge-network-status descriptor.
    :returns:
        A 7-tuple of::
            (nickname, identityDigest, descriptorDigest, timestamp,
             orAddress, orPort, dirport)
        where each value is set according to the data parsed from the
        **line**, or ``None`` if nothing suitable could be parsed.
    """
    (nickname, ID, descDigest, timestamp,
     ORaddr, ORport, dirport) = (None for x in xrange(7))

    try:
        if not line.startswith('r '):
            raise NetworkstatusParsingError(
                "Networkstatus parser received non 'r'-line: %r" % line)

        line = line[2:] # Chop off the 'r '
        fields = line.split()

        if len(fields) != 8:
            raise NetworkstatusParsingError(
                "Wrong number of fields in networkstatus 'r'-line: %r" % line)

        nickname, ID = fields[:2]

        try:
            ID = parseUnpaddedBase64(ID)
        except InvalidBase64 as error:
            raise InvalidNetworkstatusRouterIdentity(error)

        # Check the nickname validity after parsing the ID, otherwise, if the
        # nickname is invalid, we end up with the nickname being ``None`` and
        # the ID being unparsed, unpadded (meaning it is technically invalid)
        # base64.
        isValidRouterNickname(nickname)

    except NetworkstatusParsingError as error:
        logging.error(error)
        nickname, ID = None, None
    except InvalidRouterNickname as error:
        logging.error(error)
        # Assume that we mostly care about obtaining the OR's ID, then it
        # should be okay to set the nickname to ``None``, if it was invalid.
        nickname = None
    except InvalidNetworkstatusRouterIdentity as error:
        logging.error(error)
        ID = None
    else:
        try:
            descDigest = parseUnpaddedBase64(fields[2])
            timestamp = time.mktime(time.strptime(" ".join(fields[3:5]),
                                                  "%Y-%m-%d %H:%M:%S"))
            ORaddr = fields[5]
            ORport = int(fields[6])
            dirport = fields[7]
        except InvalidBase64 as error:
            logging.error(error)
            descDigest = None
        except (AttributeError, ValueError, IndexError) as error:
            logging.error(error)
            timestamp = None
    finally:
        return (nickname, ID, descDigest, timestamp, ORaddr, ORport, dirport)

@deprecate.deprecated(Version('bridgedb', 0, 2, 5))
def parseALine(line, fingerprint=None):
    """Parse an 'a'-line of a bridge networkstatus document.

    From torspec.git/dir-spec.txt, commit 36761c7d553d L1499-1512:
      |
      | "a" SP address ":" port NL
      |
      |    [Any number.]
      |
      |    Present only if the OR has at least one IPv6 address.
      |
      |    Address and portlist are as for "or-address" as specified in
      |    2.1.
      |
      |    (Only included when the vote or consensus is generated with
      |    consensus-method 14 or later.)

    :param string line: An 'a'-line from an bridge-network-status descriptor.
    :type fingerprint: string or None
    :param fingerprint: A string which identifies which OR the descriptor
        we're parsing came from (since the 'a'-line doesn't tell us, this can
        help make the log messages clearer).
    :raises: :exc:`NetworkstatusParsingError`
    :rtype: tuple
    :returns: A 2-tuple of a string respresenting the IP address and a
        :class:`bridgedb.parse.addr.PortList`.
    """
    ip = None
    portlist = None

    if line.startswith('a '):
        line = line[2:] # Chop off the 'a '
    else:
        logging.warn("Networkstatus parser received non 'a'-line for %r:"\
                     "  %r" % (fingerprint or 'Unknown', line))

    try:
        ip, portlist = line.rsplit(':', 1)
    except ValueError as error:
        logging.error("Bad separator in networkstatus 'a'-line: %r" % line)
        return (None, None)

    if ip.startswith('[') and ip.endswith(']'):
        ip = ip.strip('[]')

    try:
        if not addr.isIPAddress(ip):
            raise NetworkstatusParsingError(
                "Got invalid IP Address in networkstatus 'a'-line for %r: %r"
                % (fingerprint or 'Unknown', line))

        if addr.isIPv4(ip):
            warnings.warn(FutureWarning(
                "Got IPv4 address in networkstatus 'a'-line! "\
                "Networkstatus document format may have changed!"))
    except NetworkstatusParsingError as error:
        logging.error(error)
        ip, portlist = None, None

    try:
        portlist = addr.PortList(portlist)
        if not portlist:
            raise NetworkstatusParsingError(
                "Got invalid portlist in 'a'-line for %r!\n  Line: %r"
                % (fingerprint or 'Unknown', line))
    except (addr.InvalidPort, NetworkstatusParsingError) as error:
        logging.error(error)
        portlist = None
    else:
        logging.debug("Parsed networkstatus ORAddress line for %r:"\
                      "\n  Address: %s  \tPorts: %s"
                      % (fingerprint or 'Unknown', ip, portlist))
    finally:
        return (ip, portlist)

@deprecate.deprecated(Version('bridgedb', 0, 2, 5))
def parseSLine(line):
    """Parse an 's'-line from a bridge networkstatus document.

    The 's'-line contains all flags assigned to a bridge. The flags which may
    be assigned to a bridge are as follows:

    From torspec.git/dir-spec.txt, commit 36761c7d553d L1526-1554:
      |
      | "s" SP Flags NL
      |
      |    [Exactly once.]
      |
      |    A series of space-separated status flags, in lexical order (as ASCII
      |    byte strings).  Currently documented flags are:
      |
      |      "BadDirectory" if the router is believed to be useless as a
      |         directory cache (because its directory port isn't working,
      |         its bandwidth is always throttled, or for some similar
      |         reason).
      |      "Fast" if the router is suitable for high-bandwidth circuits.
      |      "Guard" if the router is suitable for use as an entry guard.
      |      "HSDir" if the router is considered a v2 hidden service directory.
      |      "Named" if the router's identity-nickname mapping is canonical,
      |         and this authority binds names.
      |      "Stable" if the router is suitable for long-lived circuits.
      |      "Running" if the router is currently usable.
      |      "Valid" if the router has been 'validated'.
      |      "V2Dir" if the router implements the v2 directory protocol.

    :param string line: An 's'-line from an bridge-network-status descriptor.
    :rtype: tuple
    :returns: A 2-tuple of booleans, the first is True if the bridge has the
        "Running" flag, and the second is True if it has the "Stable" flag.
    """
    line = line[2:]

    flags = [x.capitalize() for x in line.split()]
    fast    = 'Fast' in flags
    running = 'Running' in flags
    stable  = 'Stable' in flags
    guard   = 'Guard' in flags
    valid   = 'Valid' in flags

    if (fast or running or stable or guard or valid):
        logging.debug("Parsed Flags: %s%s%s%s%s"
                      % ('Fast ' if fast else '',
                         'Running ' if running else '',
                         'Stable ' if stable else '',
                         'Guard ' if guard else '',
                         'Valid ' if valid else ''))

    # Right now, we only care about 'Running' and 'Stable'
    return running, stable



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2013-2015, Isis Lovecruft
#             (c) 2007-2015, all entities within the AUTHORS file
# :license: 3-clause BSD, see included LICENSE for information

from __future__ import unicode_literals


def _(text):
    """This is necessary because strings are translated when they're imported.
    Otherwise this would make it impossible to switch languages more than
    once.

    :returns: The **text**.
    """
    return text


# TRANSLATORS: Please do not translate the word "TYPE".
EMAIL_MISC_TEXT = {
    0: _("""\
[This is an automated message; please do not reply.]"""),
    1: _("""\
Here are your bridges:"""),
    2: _("""\
You have exceeded the rate limit. Please slow down! The minimum time between
emails is %s hours. All further emails during this time period will be ignored."""),
    3: _("""\
COMMANDs: (combine COMMANDs to specify multiple options simultaneously)"""),
    # TRANSLATORS: Please DO NOT translate the word "BridgeDB".
    4: _("Welcome to BridgeDB!"),
    # TRANSLATORS: Please DO NOT translate the words "transport" or "TYPE".
    5: _("Currently supported transport TYPEs:"),
    6: _("Hey, %s!"),
    7: _("Hello, friend!"),
    8: _("Public Keys"),
    # TRANSLATORS: This string will end up saying something like:
    # "This email was generated with rainbows, unicorns, and sparkles
    #  for alice@example.com on Friday, 09 May, 2014 at 18:59:39."
    9: _("""\
This email was generated with rainbows, unicorns, and sparkles
for %s on %s at %s."""),
}

WELCOME = {
    # TRANSLATORS: Please DO NOT translate "BridgeDB".
    # TRANSLATORS: Please DO NOT translate "Pluggable Transports".
    # TRANSLATORS: Please DO NOT translate "Tor".
    # TRANSLATORS: Please DO NOT translate "Tor Network".
    0: _("""\
BridgeDB can provide bridges with several %stypes of Pluggable Transports%s,
which can help obfuscate your connections to the Tor Network, making it more
difficult for anyone watching your internet traffic to determine that you are
using Tor.\n\n"""),

    # TRANSLATORS: Please DO NOT translate "Pluggable Transports".
    1: _("""\
Some bridges with IPv6 addresses are also available, though some Pluggable
Transports aren't IPv6 compatible.\n\n"""),

    # TRANSLATORS: Please DO NOT translate "BridgeDB".
    # TRANSLATORS: The phrase "plain-ol'-vanilla" means "plain, boring,
    # regular, or unexciting". Like vanilla ice cream. It refers to bridges
    # which do not have Pluggable Transports, and only speak the regular,
    # boring Tor protocol. Translate it as you see fit. Have fun with it.
    2: _("""\
Additionally, BridgeDB has plenty of plain-ol'-vanilla bridges %s without any
Pluggable Transports %s which maybe doesn't sound as cool, but they can still
help to circumvent internet censorship in many cases.\n\n"""),
}
"""These strings should go on the first "Welcome" email sent by the
:mod:`~bridgedb.EmailServer`, as well as on the ``index.html`` template used
by the :mod:`~bridgedb.HTTPServer`. They are used as an introduction to
explain what Tor bridges are, what bridges do, and why someone might want to
use bridges.
"""

FAQ = {
    0: _("What are bridges?"),
    1: _("""\
%s Bridges %s are Tor relays that help you circumvent censorship."""),
}

OTHER_DISTRIBUTORS = {
    0: _("I need an alternative way of getting bridges!"),
    1: _("""\
Another way to get bridges is to send an email to %s. Please note that you must
send the email using an address from one of the following email providers:
%s, %s or %s."""),
}

HELP = {
    0: _("My bridges don't work! I need help!"),
    # TRANSLATORS: Please DO NOT translate "Tor".
    1: _("""If your Tor doesn't work, you should email %s."""),
    # TRANSLATORS: Please DO NOT translate "Pluggable Transports".
    # TRANSLATORS: Please DO NOT translate "Tor Browser".
    # TRANSLATORS: Please DO NOT translate "Tor".
    2: _("""\
Try including as much info about your case as you can, including the list of
bridges and Pluggable Transports you tried to use, your Tor Browser version,
and any messages which Tor gave out, etc."""),
}

BRIDGES = {
    0: _("Here are your bridge lines:"),
    1: _("Get Bridges!"),
}

OPTIONS = {
    0: _("Please select options for bridge type:"),
    1: _("Do you need IPv6 addresses?"),
    2: _("Do you need a %s?"),
}

CAPTCHA = {
    0: _('Your browser is not displaying images properly.'),
    1: _('Enter the characters from the image above...'),
}

HOWTO_TBB = {
    0: _("""How to start using your bridges"""),
    # TRANSLATORS: Please DO NOT translate "Tor Browser".
    1: _("""\
To enter bridges into Tor Browser, first go to the %s Tor Browser download
page %s and then follow the instructions there for downloading and starting
Tor Browser."""),
    # TRANSLATORS: Please DO NOT translate "Tor".
    2: _("""\
When the 'Tor Network Settings' dialogue pops up, click 'Configure' and follow
the wizard until it asks:"""),
    # TRANSLATORS: Please DO NOT translate "Tor".
    3: _("""\
Does your Internet Service Provider (ISP) block or otherwise censor connections
to the Tor network?"""),
    # TRANSLATORS: Please DO NOT translate "Tor".
    4: _("""\
Select 'Yes' and then click 'Next'. To configure your new bridges, copy and
paste the bridge lines into the text input box. Finally, click 'Connect', and
you should be good to go! If you experience trouble, try clicking the 'Help'
button in the 'Tor Network Settings' wizard for further assistance."""),
}

EMAIL_COMMANDS = {
    "get help":             _("Displays this message."),
# TRANSLATORS: Please try to make it clear that "vanilla" here refers to the
# same non-Pluggable Transport bridges described above as being
# "plain-ol'-vanilla" bridges.
    "get bridges":          _("Request vanilla bridges."),
    "get ipv6":             _("Request IPv6 bridges."),
    # TRANSLATORS: Please DO NOT translate the word the word "TYPE".
    "get transport [TYPE]": _("Request a Pluggable Transport by TYPE."),
    # TRANSLATORS: Please DO NOT translate "BridgeDB".
    # TRANSLATORS: Please DO NOT translate "GnuPG".
    "get key":              _("Get a copy of BridgeDB's public GnuPG key."),
    #"subscribe":            _("Subscribe to receive new bridges once per week"),
    #"unsubscribe":          _("Cancel a subscription to new bridges"),
}

#-----------------------------------------------------------------------------
#           All of the following containers are untranslated!
#-----------------------------------------------------------------------------

#: A list of all currently available pluggable transports. By "currently
#: available" we mean:
#:
#:   1. The PT is in a widely accepted, usable state for most Tor users.
#:   2. The PT is currently publicly deployed *en masse*".
#:   3. The PT is included within the transports which Tor Browser offers in
#:      the stable releases.
#:
CURRENT_TRANSPORTS = [
    "obfs2",
    "obfs3",
    "obfs4",
    "scramblesuit",
    "fte",
]

EMAIL_SPRINTF = {
    # Goes into the "%s types of Pluggable Transports %s" part of ``WELCOME[0]``
    "WELCOME0": ("", "[0]"),
    # Goes into the "%s without Pluggable Transport %s" part of ``WELCOME[2]``
    "WELCOME2": ("-", "-"),
    # For the "%s Tor Browser download page %s" part of ``HOWTO_TBB[1]``
    "HOWTO_TBB1": ("", "[0]"),
    # For the "you should email %s" in ``HELP[0]``
    "HELP0": ("help@rt.torproject.org"),
}
"""``EMAIL_SPRINTF`` is a dictionary that maps translated strings which
contain format specifiers (i.e. ``%s``) to what those format specifiers should
be replaced with in a given template system.

For example, a string which needs a pair of HTML ``("<a href=''">, "</a>")``
tags (for the templates used by :mod:`bridgedb.HTTPServer`) would need some
alternative replacements for the :mod:`EmailServer`, because the latter uses
templates with a ``text/plain`` mimetype instead of HTML. For the
``EmailServer``, the format strings specifiers are replaced with an empty
string where the opening ``<a>`` tags would go, and a numbered Markdown link
specifier where the closing ``</a>`` tags would go.

The keys in this dictionary are the Python variable names of the corresponding
strings which are being formatted, i.e. ``WELCOME0`` would be the string
replacements for ``strings.WELCOME.get(0)``.


For example, the ``0`` string in :data:`WELCOME` above has the substring::

    "%s without Pluggable Transport %s"

and so to replace the two ``%s`` format specifiers, you would use this mapping
like so::

>>> from bridgedb import strings
>>> welcome = strings.WELCOME[0] % strings.EMAIL_SPRINTF["WELCOME0"]
>>> print welcome.split('\n')[0]
BridgeDB can provide bridges with several types of Pluggable Transports[0],

"""

EMAIL_REFERENCE_LINKS = {
    "WELCOME0": "[0]: https://www.torproject.org/docs/pluggable-transports.html",
    "HOWTO_TBB1": "[0]: https://www.torproject.org/projects/torbrowser.html",
}



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_email_server -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Nick Mathewson <nickm@torproject.org>
#           Isis Lovecruft <isis@torproject.org> 0xA3ADB67A2CDB8B35
#           Matthew Finkel <sysrqb@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2007-2015, The Tor Project, Inc.
#             (c) 2013-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________


"""
.. py:module:: bridgedb.email.server
    :synopsis: Servers which interface with clients and distribute bridges
               over SMTP.

bridgedb.email.server
=====================

Servers which interface with clients and distribute bridges over SMTP.

::

  bridgedb.email.server
   | |_ addServer - Set up a SMTP server which listens on the configured
   |                EMAIL_PORT for incoming connections, and responds as
   |                necessary to requests for bridges.
   |
   |_ MailServerContext - Helper object that holds information used by the
   |                      email subsystem.
   |_ SMTPMessage - Plugs into Twisted Mail and receives an incoming message.
   |_ SMTPIncomingDelivery - Plugs into SMTPIncomingServerFactory and handles
   |                         SMTP commands for incoming connections.
   |_ SMTPIncomingDeliveryFactory - Factory for SMTPIncomingDeliverys.
   |_ SMTPIncomingServerFactory - Plugs into twisted.mail.smtp.SMTPFactory;
                                  creates a new SMTPMessageDelivery, which
                                  handles response email automation, whenever
                                  we get a incoming connection on the SMTP port.
..
"""

from __future__ import unicode_literals

import logging
import io
import socket

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.error import CannotListenError
from twisted.internet.task import LoopingCall
from twisted.mail import smtp
from twisted.mail.smtp import rfc822date
from twisted.python import failure

from zope.interface import implements

from bridgedb import __version__
from bridgedb import safelog
from bridgedb.crypto import initializeGnuPG
from bridgedb.email import autoresponder
from bridgedb.email import templates
from bridgedb.email import request
from bridgedb.parse import addr
from bridgedb.parse.addr import UnsupportedDomain
from bridgedb.parse.addr import canonicalizeEmailDomain


class MailServerContext(object):
    """Helper object that holds information used by email subsystem.

    :ivar str username: Reject any RCPT TO lines that aren't to this
        user. See the ``EMAIL_USERNAME`` option in the config file.
        (default: ``'bridges'``)
    :ivar int maximumSize: Reject any incoming emails longer than
        this size (in bytes). (default: 3084 bytes).
    :ivar int smtpPort: The port to use for outgoing SMTP.
    :ivar str smtpServer: The IP address to use for outgoing SMTP.
    :ivar str smtpFromAddr: Use this address in the raw SMTP ``MAIL FROM``
        line for outgoing mail. (default: ``bridges@torproject.org``)
    :ivar str fromAddr: Use this address in the email ``From:``
        line for outgoing mail. (default: ``bridges@torproject.org``)
    :ivar int nBridges: The number of bridges to send for each email.
    :ivar list blacklist: A list of blacklisted email addresses, taken from
        the ``EMAIL_BLACKLIST`` config setting.
    :ivar int fuzzyMatch: An integer specifying the maximum Levenshtein
        Distance from an incoming email address to a blacklisted email address
        for the incoming email to be dropped.
    :ivar gpg: A :class:`gnupg.GPG` interface_, as returned by
        :func:`~bridgedb.crypto.initialiseGnuPG`, or ``None`` if we couldn't
        initialize GnuPG for some reason.
    :ivar gpgSignFunc: A callable which signs a message, e.g. the one returned
        from :func:`~bridgedb.crypto.initialiseGnuPG`.
    """

    def __init__(self, config, distributor, schedule):
        """Create a context for storing configs for email bridge distribution.

        :type config: :class:`bridgedb.persistent.Conf`
        :type distributor: :class:`bridgedb.Dist.EmailBasedDistributor`
        :param distributor: The distributor will handle getting the correct
            bridges (or none) for a client for us.
        :type schedule: :class:`bridgedb.schedule.ScheduledInterval`
        :param schedule: An interval-based scheduler, used to help the
            :data:`distributor` know if we should give bridges to a client.
        """
        self.config = config
        self.distributor = distributor
        self.schedule = schedule

        self.maximumSize = smtp.SMTP.MAX_LENGTH
        self.includeFingerprints = config.EMAIL_INCLUDE_FINGERPRINTS
        self.nBridges = config.EMAIL_N_BRIDGES_PER_ANSWER

        self.username = (config.EMAIL_USERNAME or "bridges")
        self.hostname = socket.gethostname()
        self.fromAddr = (config.EMAIL_FROM_ADDR or "bridges@torproject.org")
        self.smtpFromAddr = (config.EMAIL_SMTP_FROM_ADDR or self.fromAddr)
        self.smtpServerPort = (config.EMAIL_SMTP_PORT or 25)
        self.smtpServerIP = (config.EMAIL_SMTP_HOST or "127.0.0.1")

        self.domainRules = config.EMAIL_DOMAIN_RULES or {}
        self.domainMap = config.EMAIL_DOMAIN_MAP or {}
        self.canon = self.buildCanonicalDomainMap()
        self.whitelist = config.EMAIL_WHITELIST or {}
        self.blacklist = config.EMAIL_BLACKLIST or []
        self.fuzzyMatch = config.EMAIL_FUZZY_MATCH or 0

        self.gpg, self.gpgSignFunc = initializeGnuPG(config)

    def buildCanonicalDomainMap(self):
        """Build a map for all email provider domains from which we will accept
        emails to their canonical domain name.

        .. note:: Be sure that ``MailServerContext.domainRules`` and
            ``MailServerContext.domainMap`` are set appropriately before calling
            this method.

        This method is automatically called during initialisation, and the
        resulting domain map is stored as ``MailServerContext.canon``.

        :rtype: dict
        :returns: A dictionary which maps all domains and subdomains which we
            accept emails from to their second-level, canonical domain names.
        """
        canon = self.domainMap
        for domain, rule in self.domainRules.items():
            if domain not in canon.keys():
                canon[domain] = domain
        for domain in self.config.EMAIL_DOMAINS:
            canon[domain] = domain
        return canon


class SMTPMessage(object):
    """Plugs into the Twisted Mail and receives an incoming message.

    :var list lines: A list of lines from an incoming email message.
    :var int nBytes: The number of bytes received thus far.
    :var bool ignoring: If ``True``, we're ignoring the rest of this message
        because it exceeded :data:`MailServerContext.maximumSize`.
    :var canonicalFromSMTP: See
        :meth:`~bridgedb.email.autoresponder.SMTPAutoresponder.runChecks`.
    :var canonicalFromEmail: See
        :meth:`~bridgedb.email.autoresponder.SMTPAutoresponder.runChecks`.
    :var canonicalDomainRules: See
        :meth:`~bridgedb.email.autoresponder.SMTPAutoresponder.runChecks`.
    :var message: (:api:`twisted.mail.smtp.rfc822.Message` or ``None``) The
        incoming email message.
    :var responder: A :class:`~bridgedb.email.autoresponder.SMTPAutoresponder`
        which parses and checks the incoming :data:`message`. If it decides to
        do so, it will build a
        :meth:`~bridgedb.email.autoresponder.SMTPAutoresponder.reply` email
        and :meth:`~bridgedb.email.autoresponder.SMTPAutoresponder.send` it.
    """
    implements(smtp.IMessage)

    def __init__(self, context, canonicalFromSMTP=None):
        """Create a new SMTPMessage.

        These are created automatically via
        :class:`SMTPIncomingDelivery`.

        :param context: The configured :class:`MailServerContext`.
        :type canonicalFromSMTP: str or None
        :param canonicalFromSMTP: The canonical domain which this message was
            received from. For example, if ``'gmail.com'`` is the configured
            canonical domain for ``'googlemail.com'`` and a message is
            received from the latter domain, then this would be set to the
            former.
        """
        self.context = context
        self.canon = context.canon
        self.canonicalFromSMTP = canonicalFromSMTP
        self.canonicalFromEmail = None
        self.canonicalDomainRules = None

        self.lines = []
        self.nBytes = 0
        self.ignoring = False

        self.message = None
        self.responder = autoresponder.SMTPAutoresponder()
        self.responder.incoming = self

    def lineReceived(self, line):
        """Called when we get another line of an incoming message."""
        self.nBytes += len(line)
        if self.nBytes > self.context.maximumSize:
            self.ignoring = True
        else:
            self.lines.append(line)
        if not safelog.safe_logging:
            try:
                ln = line.rstrip("\r\n").encode('utf-8', 'replace')
                logging.debug("> %s" % ln)
            except (UnicodeError, UnicodeDecodeError):  # pragma: no cover
                pass
            except Exception as error:  # pragma: no cover
                logging.error("Error while trying to log incoming email")
                logging.exception(error)

    def eomReceived(self):
        """Tell the :data:`responder` to reply when we receive an EOM."""
        if not self.ignoring:
            self.message = self.getIncomingMessage()
            self.responder.reply()
        return defer.succeed(None)

    def connectionLost(self):
        """Called if we die partway through reading a message."""
        pass

    def getIncomingMessage(self):
        """Create and parse an :rfc:`2822` message object for all :data:`lines`
        received thus far.

        :rtype: :api:`twisted.mail.smtp.rfc822.Message`
        :returns: A ``Message`` comprised of all lines received thus far.
        """
        rawMessage = io.StringIO()
        for line in self.lines:
            rawMessage.writelines(unicode(line.decode('utf8')) + u'\n')
        rawMessage.seek(0)
        return smtp.rfc822.Message(rawMessage)


class SMTPIncomingDelivery(smtp.SMTP):
    """Plugs into :class:`SMTPIncomingServerFactory` and handles SMTP commands
    for incoming connections.

    :type context: :class:`MailServerContext`
    :var context: A context containing SMTP/Email configuration settings.
    :var deferred: A :api:`deferred <twisted.internet.defer.Deferred>` which
        will be returned when :meth:`reply` is called. Additional callbacks
        may be set on this deferred in order to schedule additional actions
        when the response is being sent.
    :type fromCanonicalSMTP: str or ``None``
    :var fromCanonicalSMTP: If set, this is the canonicalized domain name of
       the address we received from incoming connection's ``MAIL FROM:``.
    """
    implements(smtp.IMessageDelivery)

    context = None
    deferred = defer.Deferred()
    fromCanonicalSMTP = None

    @classmethod
    def setContext(cls, context):
        """Set our :data:`context` to a new :class:`MailServerContext`."""
        cls.context = context

    def receivedHeader(self, helo, origin, recipients):
        """Create the ``Received:`` header for an incoming email.

        :type helo: tuple
        :param helo: The lines received during SMTP client HELO.
        :type origin: :api:`twisted.mail.smtp.Address`
        :param origin: The email address of the sender.
        :type recipients: list
        :param recipients: A list of :api:`twisted.mail.smtp.User` instances.
        """
        helo_ = ' helo={0}'.format(helo[0]) if helo[0] else ''
        from_ = 'from %s ([%s]%s)' % (helo[0], helo[1], helo_)
        by_ = 'by %s with BridgeDB (%s)' % (smtp.DNSNAME, __version__)
        for_ = 'for %s; %s ' % (' '.join(map(str, recipients)), rfc822date())
        return str('Received: %s\n\t%s\n\t%s' % (from_, by_, for_))

    def validateFrom(self, helo, origin):
        """Validate the ``MAIL FROM:`` address on the incoming SMTP connection.

        This is done at the SMTP layer. Meaning that if a Postfix or other
        email server is proxying emails from the outside world to BridgeDB,
        the :api:`origin.domain <twisted.email.smtp.Address.domain>` will be
        set to the local hostname. Therefore, if the SMTP ``MAIL FROM:``
        domain name is our own hostname (as returned from
        :func:`socket.gethostname`) or our own FQDN, allow the connection.

        Otherwise, if the ``MAIL FROM:`` domain has a canonical domain in our
        mapping (taken from our :data:`context.canon`, which is taken in turn
        from the ``EMAIL_DOMAIN_MAP``), then our :data:`fromCanonicalSMTP` is
        set to that domain.

        :type helo: tuple
        :param helo: The lines received during SMTP client HELO.
        :type origin: :api:`twisted.mail.smtp.Address`
        :param origin: The email address we received this message from.
        :raises: :api:`twisted.mail.smtp.SMTPBadSender` if the
            ``origin.domain`` was neither our local hostname, nor one of the
            canonical domains listed in :ivar:`context.canon`.
        :rtype: :api:`twisted.mail.smtp.Address`
        :returns: The ``origin``. We *must* return some non-``None`` data from
            this method, or else Twisted will reply to the sender with a 503
            error.
        """
        try:
            if str(origin) in self.context.whitelist.keys():
                logging.warn("Got SMTP 'MAIL FROM:' whitelisted address: %s."
                             % str(origin))
                # We need to be certain later that when the fromCanonicalSMTP
                # domain is checked again the email 'From:' canonical domain,
                # that we allow whitelisted addresses through the check.
                self.fromCanonicalSMTP = origin.domain
                return origin
            if ((origin.domain == self.context.hostname) or
                (origin.domain == smtp.DNSNAME)):
                self.fromCanonicalSMTP = origin.domain
            else:
                logging.debug("Canonicalizing client SMTP domain...")
                canonical = canonicalizeEmailDomain(origin.domain,
                                                    self.context.canon)
                logging.debug("Canonical SMTP domain: %r" % canonical)
                self.fromCanonicalSMTP = canonical
        except UnsupportedDomain as error:
            logging.info(error)
            raise smtp.SMTPBadSender(origin)
        except Exception as error:
            logging.exception(error)

        # This method **cannot** return None, or it'll cause a 503 error.
        return origin

    def validateTo(self, user):
        """Validate the SMTP ``RCPT TO:`` address for the incoming connection.

        The local username and domain name to which this SMTP message is
        addressed, after being stripped of any ``'+'`` aliases, **must** be
        identical to those in the email address set our
        ``EMAIL_SMTP_FROM_ADDR`` configuration file option.

        :type user: :api:`twisted.mail.smtp.User`
        :param user: Information about the user this SMTP message was
            addressed to.
        :raises: A :api:`twisted.mail.smtp.SMTPBadRcpt` if any of the above
            conditions weren't met.
        :rtype: callable
        :returns: A parameterless function which returns an instance of
            :class:`SMTPMessage`.
        """
        logging.debug("Validating SMTP 'RCPT TO:' email address...")

        recipient = user.dest
        ourAddress = smtp.Address(self.context.smtpFromAddr)

        if not ((ourAddress.domain in recipient.domain) or
                (recipient.domain == "bridgedb")):
            logging.debug(("Not our domain (%s) or subdomain, skipping"
                           " SMTP 'RCPT TO' address: %s")
                          % (ourAddress.domain, str(recipient)))
            raise smtp.SMTPBadRcpt(str(recipient))
        # The recipient's username should at least start with ours,
        # but it still might be a '+' address.
        if not recipient.local.startswith(ourAddress.local):
            logging.debug(("Username doesn't begin with ours, skipping"
                           " SMTP 'RCPT TO' address: %s") % str(recipient))
            raise smtp.SMTPBadRcpt(str(recipient))
        # Ignore everything after the first '+', if there is one.
        beforePlus = recipient.local.split('+', 1)[0]
        if beforePlus != ourAddress.local:
            raise smtp.SMTPBadRcpt(str(recipient))

        return lambda: SMTPMessage(self.context, self.fromCanonicalSMTP)


class SMTPIncomingDeliveryFactory(object):
    """Factory for :class:`SMTPIncomingDelivery` s.

    This class is used to distinguish between different messages delivered
    over the same connection. This can be used to optimize delivery of a
    single message to multiple recipients, something which cannot be done by
    :api:`IMessageDelivery <twisted.mail.smtp.IMessageDelivery>` implementors
    due to their lack of information.

    :var context: A :class:`MailServerContext` for storing configuration settings.
    :var delivery: A :class:`SMTPIncomingDelivery` to deliver incoming
        SMTP messages to.
    """
    implements(smtp.IMessageDeliveryFactory)

    context = None
    delivery = SMTPIncomingDelivery

    def __init__(self):
        logging.debug("%s created." % self.__class__.__name__)

    @classmethod
    def setContext(cls, context):
        """Set our :data:`context` and the context for our :data:`delivery`."""
        cls.context = context
        cls.delivery.setContext(cls.context)

    def getMessageDelivery(self):
        """Get a new :class:`SMTPIncomingDelivery` instance."""
        return self.delivery()


class SMTPIncomingServerFactory(smtp.SMTPFactory):
    """Plugs into :api:`twisted.mail.smtp.SMTPFactory`; creates a new
    :class:`SMTPIncomingDeliveryFactory`, which handles response email
    automation whenever we get a incoming connection on the SMTP port.

    .. warning:: My :data:`context` isn't an OpenSSL context, as is used for
        the :api:`twisted.mail.smtp.ESMTPSender`.

    :ivar context: A :class:`MailServerContext` for storing configuration settings.
    :ivar deliveryFactory: A :class:`SMTPIncomingDeliveryFactory` for
        producing :class:`SMTPIncomingDelivery`s.
    :ivar domain: :api:`Our FQDN <twisted.mail.smtp.DNSNAME>`.
    :ivar int timeout: The number of seconds to wait, after the last chunk of
        data was received, before raising a
        :api:`SMTPTimeoutError <twisted.mail.smtp.SMTPTimeoutError>` for an
        incoming connection.
    :ivar protocol: :api:`SMTP <twisted.mail.smtp.SMTP>`
    """

    context = None
    deliveryFactory = SMTPIncomingDeliveryFactory

    def __init__(self, **kwargs):
        smtp.SMTPFactory.__init__(self, **kwargs)
        self.deliveryFactory = self.deliveryFactory()

    @classmethod
    def setContext(cls, context):
        """Set :data:`context` and :data:`deliveryFactory`.context."""
        cls.context = context
        cls.deliveryFactory.setContext(cls.context)

    def buildProtocol(self, addr):
        p = smtp.SMTPFactory.buildProtocol(self, addr)
        self.deliveryFactory.transport = p.transport # XXX is this set yet?
        p.factory = self
        p.deliveryFactory = self.deliveryFactory
        return p


def addServer(config, distributor, schedule):
    """Set up a SMTP server which listens on the configured ``EMAIL_PORT`` for
    incoming connections, and responds as necessary to requests for bridges.

    :type config: :class:`bridgedb.configure.Conf`
    :param config: A configuration object.
    :type distributor: :class:`bridgedb.Dist.EmailBasedDistributor`
    :param dist: A distributor which will handle database interactions, and
        will decide which bridges to give to who and when.
    :type schedule: :class:`bridgedb.schedule.ScheduledInterval`
    :param schedule: The schedule. XXX: Is this even used?
    """
    context = MailServerContext(config, distributor, schedule)
    factory = SMTPIncomingServerFactory()
    factory.setContext(context)

    addr = config.EMAIL_BIND_IP or ""
    port = config.EMAIL_PORT or 6725

    try:
        reactor.listenTCP(port, factory, interface=addr)
    except CannotListenError as error:  # pragma: no cover
        logging.fatal(error)
        raise SystemExit(error.message)

    # Set up a LoopingCall to run every 30 minutes and forget old email times.
    lc = LoopingCall(distributor.cleanDatabase)
    lc.start(1800, now=False)

    return factory
